# This code is heavily based upon winioport.py
# Provides hardware port access for Python under Windows 95/98/NT/2000
#
# Original Author: Dincer Aydin dinceraydin@gmx.net www.dinceraydin.com
# Merged directly into psychopy by: Mark Hymers <mark.hymers@ynic.york.ac.uk>
# All bugs are Mark's fault.
#
# This module depends on:
#   ctypes Copyright (c) 2000, 2001, 2002, 2003 Thomas Heller
#   DLPortIO Win32 DLL hardware I/O functions & Kernel mode driver for WinNT
#
# In this package you will find almost any sort of port IO function one may
# imagine. Values of port registers are srored in temporary variables. This is
# for the bit set/reset functions to work right Some register bits are inverted.
# on the port pins, but you need not worry about them. The functions in this
# module take this into account. For eaxample when you call
# winioport.pportDataStrobe(1) the data strobe pin of the printer port will go
# HIGH.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files , to deal in the Software
# without restriction, including without limitation the rights to use, copy,
# modify, merge, publish,and distribute copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following
# conditions:
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

from __future__ import absolute_import, print_function, division
from builtins import *

class PParallelDLPortIO(object):
    """
    This class provides read/write access to the parallel port on a PC.

    This is a wrapper around Dincer Aydin's `winioport`_ for reading and writing to
    the parallel port, but adds the following additional functions for convenience.

    On windows `winioport`_ requires the `PortIO driver`_ to be installed.

    An alternative on Linux might be to use PyParallel
    An alternative on other versions of Windows might be to use inpout32.

    .  _winioport: http://www.dinceraydin.com/python/indexeng.html
    .. _PortIO driver: http://www.winfordeng.com/support/download.php
    """

    def __init__(self, address=0x0378):
        """Set the memory address of your parallel port, to be used in
        subsequent method calls on this class.

        common port addresses::

            LPT1 = 0x0378 or 0x03BC
            LPT2 = 0x0278 or 0x0378
            LPT3 = 0x0278
        """

        from ctypes import windll
        try:
            # Load dlportio.dll functions
            self.port = windll.dlportio
        except Exception as e:
            print("Could not import DLportIO driver, parallel Ports not available")
            raise e

        if isinstance(address, str) and address.startswith('0x'): #convert u"0x0378" into 0x0378
            self.base = int(address, 16)
        else:
            self.base = address
        self.reverse = False
        self.status = None

    @property
    def reverse(self):
        """Getter for reverse."""

        return self._reverse

    @reverse.setter
    def reverse(self, value):
        """Setter for reverse."""

        if value:
            self.port.DlPortWritePortUchar(self.base + 2, int(self.port.DlPortReadPortUchar(self.base + 2) | ((1 << 5) & 255)))
            self._reverse = True
        else:
            self.port.DlPortWritePortUchar(self.base + 2, int(self.port.DlPortReadPortUchar(self.base + 2) & (~(1 << 5) & 255)))
            self._reverse = False

    def setData(self, data):
        """Set the data to be presented on the parallel port (one ubyte).
        Alternatively you can set the value of each pin (data pins are pins
        2-9 inclusive) using :func:`setPin`

        examples::

           p.setData(0) #sets all pins low
           p.setData(255) #sets all pins high
           p.setData(2) #sets just pin 3 high (remember that pin2=bit0)
           p.setData(3) #sets just pins 2 and 3 high

        you can also convert base 2 to int v easily in python::

           p.setData( int("00000011",2) )#pins 2 and 3 high
           p.setData( int("00000101",2) )#pins 2 and 4 high

        """
        self.port.DlPortWritePortUchar(self.base, data)

    def setPin(self, pinNumber, state):
        """Set a desired pin to be high(1) or low(0).

        Only pins 2-9 (incl) are normally used for data output::

            p.setPin(3, 1)#sets pin 3 high
            p.setPin(3, 0)#sets pin 3 low
        """
        # I can't see how to do this without reading and writing the data
        # or caching the registers which seems like a very bad idea...
        if state:
            if pinNumber in range(2, 10):
                self.port.DlPortWritePortUchar( self.base, self.port.DlPortReadPortUchar( self.base ) | (2**(pinNumber-2)) )
            elif pinNumber==1:
                self.port.DlPortWritePortUchar( self.base + 2, self.port.DlPortReadPortUchar( self.base + 2) | 1 )
            elif pinNumber==14:
                self.port.DlPortWritePortUchar( self.base + 2, self.port.DlPortReadPortUchar( self.base + 2) | 2 )
            elif pinNumber==16:
                self.port.DlPortWritePortUchar( self.base + 2, self.port.DlPortReadPortUchar( self.base + 2) | 4 )
            elif pinNumber==17:
                self.port.DlPortWritePortUchar( self.base + 2, self.port.DlPortReadPortUchar( self.base + 2) | 8 )
        else:
            if pinNumber in range(2, 10):
                self.port.DlPortWritePortUchar( self.base, self.port.DlPortReadPortUchar( self.base ) & (255 ^ 2**(pinNumber-2)) )
            elif pinNumber==1:
                self.port.DlPortWritePortUchar( self.base + 2, self.port.DlPortReadPortUchar( self.base + 2 ) & (255 ^ 1) )
            elif pinNumber==14:
                self.port.DlPortWritePortUchar( self.base + 2, self.port.DlPortReadPortUchar( self.base + 2 ) & (255 ^ 2) )
            elif pinNumber==16:
                self.port.DlPortWritePortUchar( self.base + 2, self.port.DlPortReadPortUchar( self.base + 2 ) & (255 ^ 4) )
            elif pinNumber==17:
                self.port.DlPortWritePortUchar( self.base + 2, self.port.DlPortReadPortUchar( self.base + 2 ) & (255 ^ 8) )

    def readData(self):
        """Return the value currently set on the data pins (2-9)"""
        return (self.port.DlPortReadPortUchar( self.base ))

    def readPin(self, pinNumber):
        """Determine whether a desired (input) pin is high(1) or low(0).

        Pins 1-17 are currently read here
        """
        if pinNumber==10:
            return (self.port.DlPortReadPortUchar( self.base + 1 ) >> 6) & 1 # 10 = ACK
        elif pinNumber==11:
            return (self.port.DlPortReadPortUchar( self.base + 1 ) >> 7) & 1 # 11 = BUSY
        elif pinNumber==12:
            return (self.port.DlPortReadPortUchar( self.base + 1 ) >> 5) & 1 # 12 = PAPER-OUT
        elif pinNumber==13:
            return (self.port.DlPortReadPortUchar( self.base + 1 ) >> 4) & 1 # 13 = SELECT
        elif pinNumber==15:
            return (self.port.DlPortReadPortUchar( self.base + 1 ) >> 3) & 1 # 15 = ERROR
        elif pinNumber >= 2 and pinNumber <= 9:
            return (self.port.DlPortReadPortUchar( self.base ) >> (pinNumber - 2)) & 1
        elif pinNumber==1:
            return (self.port.DlPortReadPortUchar( self.base + 2 ) >> 0) & 1 # STROBE
        elif pinNumber==14:
            return (self.port.DlPortReadPortUchar( self.base + 2 ) >> 1) & 1 # LineFeed
        elif pinNumber==16:
            return (self.port.DlPortReadPortUchar( self.base + 2 ) >> 2) & 1 # RESET
        elif pinNumber==17:
            return (self.port.DlPortReadPortUchar( self.base + 2 ) >> 3) & 1 # Select Printer
        else:
            print('Pin %i cannot be read (by the PParallelDLPortIO.readPin() yet)' % (pinNumber))




"""
A streaming button box.

This module contains a class implementing a streaming button box.

"""
from __future__ import absolute_import, print_function, division
from builtins import *

__author__ = 'Florian Krause <florian@expyriment.org>, \
Oliver Lindemann <oliver@expyriment.org>'
__version__ = ''
__revision__ = ''
__date__ = ''


from types import FunctionType

from . import defaults
from .. import _internals
from ..misc import compare_codes
from .._internals import CallbackQuitEvent, skip_wait_methods
from ..misc._timer import get_time
from ._input_output import Input, Output


class StreamingButtonBox(Input, Output):
    """A class implementing a streaming button box input."""

    def __init__(self, interface, baseline):
        """Create a streaming button box input.

        Parameters
        ----------
        interface : ``expyriment.io.SerialPort`` or ``expyriment.io.ParallelPort``
            the interface to use
        baseline : int
            code that is sent when nothing is pressed (int)

        """

        Input.__init__(self)
        Output.__init__(self)
        self._interface = interface
        if baseline is not None:
            self._baseline = baseline
        else:
            self._baseline = defaults.streamingbuttonbox_baseline

    @property
    def interface(self):
        """Getter for interface."""
        return self._interface

    @property
    def baseline(self):
        """Getter for baseline."""
        return self._baseline

    @baseline.setter
    def baseline(self, value):
        """Setter for baseline"""
        self._baseline = value

    def clear(self):
        """Clear the receive buffer (if available)."""

        self._interface.clear()
        if self._logging:
            _internals.active_exp._event_file_log("{0},cleared".format(
            self.__class__.__name__), 2)

    def check(self, codes=None, bitwise_comparison=False):
        """Check for response codes.

        If bitwise_comparison = True, the function performs a bitwise
        comparison (logical and) between codes and received input.

        Parameters
        ----------
        codes : int or list, optional
            certain bit pattern or list of bit pattern to wait for,
            if codes is not set (None) the function returns
            for any event that differs from the baseline
        bitwise_comparison : bool, optional
            make a bitwise comparison (default=False)

        Returns
        -------
        key : int
            key code or None

        """

        while True:
            read = self._interface.poll()
            if read is not None:
                if codes is None and read != self._baseline:
                    if self._logging:
                        _internals.active_exp._event_file_log(
                        "{0},received,{1},check".format(
                            self.__class__.__name__,
                            read), 2)
                    return read
                elif compare_codes(read, codes, bitwise_comparison):
                    if self._logging:
                        _internals.active_exp._event_file_log(
                        "{0},received,{1},check".format(
                            self.__class__.__name__,
                            read))
                    return read
            else:
                return None

    def wait(self, codes=None, duration=None, no_clear_buffer=False,
             bitwise_comparison=False, callback_function=None,
             process_control_events=True):
        """Wait for responses defined as codes.

        If bitwise_comparison = True, the function performs a bitwise
        comparison (logical and) between codes and received input and waits
        until a certain bit pattern is set.

        Parameters
        ----------
        codes : int or list, optional
            bit pattern to wait for
            if codes is not set (None) the function returns for any
            event that differs from the baseline
        duration : int, optional
            maximal time to wait in ms
        no_clear_buffer : bool, optional
            do not clear the buffer (default = False)
        bitwise_comparison : bool, optional
            make a bitwise comparison (default = False)
        callback_function : function, optional
            function to repeatedly execute during waiting loop
        process_control_events : bool, optional
            process ``io.Keyboard.process_control_keys()`` and
            ``io.Mouse.process_quit_event()`` (default = True)

        Returns
        -------
        key : int
            key code (or None) that quited waiting
        rt : int
            reaction time

        Notes
        -----
        This will also by default process control events (quit and pause).
        Thus, keyboard events will be cleared from the cue and cannot be
        received by a Keyboard().check() anymore!

        See Also
        --------
        expyriment.design.Experiment.register_wait_callback_function

        """

        if _internals.skip_wait_methods:
            return None, None
        start = get_time()
        rt = None
        if not no_clear_buffer:
            self.clear()
        while True:
            if isinstance(callback_function, FunctionType):
                rtn_callback = callback_function()
                if isinstance(rtn_callback, CallbackQuitEvent):
                    found = rtn_callback
                    rt = int((get_time() - start) * 1000)
                    break
            if _internals.active_exp.is_initialized:
                rtn_callback = _internals.active_exp._execute_wait_callback()
                if isinstance(rtn_callback, CallbackQuitEvent):
                    found = rtn_callback
                    rt = int((get_time() - start) * 1000)
                    break
                if process_control_events:
                    if _internals.active_exp.mouse.process_quit_event() or \
                       _internals.active_exp.keyboard.process_control_keys():
                        break
                else:
                    _internals.pump_pygame_events()
            if duration is not None:
                if int((get_time() - start) * 1000) > duration:
                    return None, None
            found = self.check(codes, bitwise_comparison)
            if found is not None:
                rt = int((get_time() - start) * 1000)
                break

        if self._logging:
            _internals.active_exp._event_file_log(
                                "{0},received,{1},wait".format(
                                                self.__class__.__name__,
                                                found))
        return found, rt



#!/usr/bin/env python
"""
Setup file for Expyriment
"""


from __future__ import print_function
from builtins import *


__author__ = 'Florian Krause <florian@expyriment.org>, \
Oliver Lindemann <oliver@expyriment.org>'


import stat
import os
import sys
from subprocess import Popen, PIPE, call
try:
    from setuptools import setup
    from setuptools.command.sdist import sdist
    from setuptools.command.build_py import build_py
    from setuptools.command.install import install
    from setuptools.command.install_data import install_data
    from setuptools.command.bdist_wininst import bdist_wininst
except ImportError:
    from distutils.core import setup
    from distutils.command.sdist import sdist
    from distutils.command.build_py import build_py
    from distutils.command.install import install
    from distutils.command.install_data import install_data
    from distutils.command.bdist_wininst import bdist_wininst
from os import remove, close, chmod, path
from shutil import move, copytree, rmtree
from tempfile import mkstemp
from glob import glob


# Settings
description='A Python library for cognitive and neuroscientific experiments'
author='Florian Krause, Oliver Lindemann'
author_email='florian@expyriment.org, oliver@expyriment.org'
license='GNU GPLv3'
url='http://www.expyriment.org'

package_dir={'expyriment': 'expyriment'}

packages = ['expyriment',
            'expyriment.control',
            'expyriment.io', 'expyriment.io.extras',
            'expyriment.io._parallelport',
            'expyriment.misc', 'expyriment.misc.extras',
            'expyriment.misc.data_preprocessing',
            'expyriment.misc.geometry',
            'expyriment.misc.statistics',
            'expyriment.stimuli', 'expyriment.stimuli.extras',
            'expyriment.design', 'expyriment.design.extras',
            'expyriment.design.randomize',
            'expyriment.design.permute']

package_data = {'expyriment': ['expyriment_logo.png', '_fonts/*.*']}

data_files = [('share/expyriment/documentation/api',
               glob('documentation/api/*.*')),
              ('share/expyriment/documentation/sphinx',
               glob('documentation/sphinx/*.*')),
              ('share/expyriment/documentation/sphinx',
               glob('documentation/sphinx/Makefile'))]

source_files = ['.release_info',
                'CHANGES.MD',
                'COPYING.txt',
                'Makefile',
                'README.md']

install_requires = ["future>=0.15,<1",
                    "pygame>=1.9,<2",
                    "pyopengl>=3.0,<4"]

extras_require = {
    'data_preprocessing': ["numpy>=1.6,<2"],
    'serialport':         ["pyserial>=3,<4"],
    'parallelport_linux': ["pyparallel>=0.2,<1"],
    'video':              ["sounddevice>=0.3,<1",
                           "mediadecoder>=0.1,<1"],
    'all':                ["numpy>=1.6,<2",
                           "pyserial>=3,<4",
                           "pyparallel>=0.2,<1",
                           "sounddevice>=0.3,<1",
                           "mediadecoder>=0.1,<1"],
    }

entry_points = {
        'console_scripts': ['expyriment=expyriment.cli:main'],
    }


class Sdist(sdist):
    def get_file_list(self):
        version_nr, revision_nr, date = get_version_info_from_release_info()
        # If code is check out from GitHub repository, change .release_info
        if date.startswith("$Format:"):
            version_nr, revision_nr, date = get_version_info_from_git()
            version_nr = "tag: " + version_nr
            move(".release_info", ".release_info.bak")
            with open(".release_info", 'w') as f:
                f.write(u"{0}\n{1}\n{2}".format(version_nr, revision_nr, date))
        for f in source_files:
            self.filelist.append(f)
        sdist.get_file_list(self)

    def run(self):
        sdist.run(self)
        try:
            move(".release_info.bak", ".release_info")
        except:
            pass


# Manipulate the header of all files (only for building/installing from
# repository)
class Build(build_py):
    """Specialized Python source builder."""

    def byte_compile(self, files):
        for f in files:
            if f.endswith('.py'):
                # Create temp file
                fh, abs_path = mkstemp()
                new_file = open(abs_path, 'wb')
                old_file = open(f, 'rb')  # was 'rUb'
                for line in old_file:
                    if line[0:11] == b'__version__':
                        new_file.write("__version__ = '{0}'\n".format(
                            version_nr).encode("utf-8"))
                    elif line[0:12] == b'__revision__':
                        new_file.write("__revision__ = '{0}'\n".format(
                            revision_nr).encode("utf-8"))
                    elif line[0:8] == b'__date__':
                        new_file.write("__date__ = '{0}'\n".format(
                            date).encode("utf-8"))
                    else:
                        new_file.write(line)
                # Close temp file
                new_file.close()
                close(fh)
                old_file.close()
                # Remove original file
                remove(f)
                # Move new file
                move(abs_path, f)
                chmod(f,
                      stat.S_IRUSR | stat.S_IWUSR | stat.S_IRGRP | stat.S_IROTH)
        build_py.byte_compile(self, files)


# Clear old installation when installing
class Install(install):
    """Specialized installer."""

    def run(self):
        # Clear old installation
        try:
            olddir = path.abspath(self.install_lib + path.sep + "expyriment")
            oldegginfo = glob(path.abspath(self.install_lib) + path.sep +
                              "expyriment*.egg-info")
            for egginfo in oldegginfo:
                remove(egginfo)
            if path.isdir(olddir):
                rmtree(olddir)
        except:
            pass
        install.run(self)


# Clear old installation when installing (for bdist_wininst)
class Wininst(bdist_wininst):
    """Specialized installer."""

    def run(self):
        fh, abs_path = mkstemp(".py")
        new_file = open(abs_path, 'w')
        # Clear old installation
        new_file.write("""
from distutils import sysconfig
import os, shutil
old_installation = os.path.join(sysconfig.get_python_lib(), 'expyriment')
if os.path.isdir(old_installation):
    shutil.rmtree(old_installation)
""")
        new_file.close()
        close(fh)
        self.pre_install_script = abs_path
        bdist_wininst.run(self)


# Build Sphinx HTML documentation and add them to data_files
class InstallData(install_data):
    def run(self):

        # Try to build/add documentation
        try:
            cwd = os.getcwd()
            copytree('expyriment', 'documentation/sphinx/expyriment')
            os.chdir('documentation/sphinx/')
            call([sys.executable, "./create_rst_api_reference.py"])
            call(["sphinx-build", "-b", "html", "-d", "_build/doctrees", ".", "_build/html"])
            os.chdir(cwd)
            self.data_files.append(('share/expyriment/documentation/html',
                               glob('documentation/sphinx/_build/html/*.*')))
            self.data_files.append(('share/expyriment/documentation/html/_downloads',
                               glob('documentation/sphinx/_build/html/_downloads/*.*')))
            self.data_files.append(('share/expyriment/documentation/html/_images',
                               glob('documentation/sphinx/_build/html/_images/*.*')))
            self.data_files.append(('share/expyriment/documentation/html/_sources',
                               glob('documentation/sphinx/_build/html/_sources/*.*')))
            self.data_files.append(('share/expyriment/documentation/html/_static',
                               glob('documentation/sphinx/_build/html/_static/*.*')))
            self.data_files.append(('share/expyriment/documentation/html/_static/css',
                               glob('documentation/sphinx/_build/html/_static/css/*.*')))
            self.data_files.append(('share/expyriment/documentation/html/_static/fonts',
                               glob('documentation/sphinx/_build/html/_static/fonts/*.*')))
            self.data_files.append(('share/expyriment/documentation/html/_static/js',
                               glob('documentation/sphinx/_build/html/_static/js/*.*')))
        except:
            html_created = False
            warning = "HTML documentation NOT created! (sphinx and numpydoc installed?)"
            os.chdir(cwd)

        # Install data
        install_data.run(self)

        # Clean up Sphinx folder
        rmtree("documentation/sphinx/expyriment", ignore_errors=True)
        rmtree("documentation/sphinx/_build", ignore_errors=True)
        for file_ in glob("documentation/sphinx/expyriment.*"):
            remove_file(file_)
        remove_file("documentation/sphinx/Changelog.rst")
        #remove_file("documentation/sphinx/CommandLineInterface.rst")
        remove_file("documentation/sphinx/sitemap.yml")


# Helper functions
def remove_file(file_):
    try:
        os.remove(file_)
    except:
        pass

def get_version_info_from_git():
    """Get version number, revision number and date from git repository."""

    proc = Popen(['git', 'describe', '--tags', '--dirty', '--always'], \
                        stdout=PIPE, stderr=PIPE)
    version_nr = "{0}".format(
        proc.stdout.read().lstrip(b"v").strip().decode("utf-8"))
    proc = Popen(['git', 'log', '--format=%H', '-1'], \
                        stdout=PIPE, stderr=PIPE)
    revision_nr = proc.stdout.read().strip()[:7].decode("utf-8")
    proc = Popen(['git', 'log', '--format=%cd', '-1'],
                     stdout=PIPE, stderr=PIPE)
    date = proc.stdout.readline().strip().decode("utf-8")
    return version_nr, revision_nr, date

def get_version_info_from_release_info():
    """Get version number, revision number and date from .release_info."""

    setup_dir = os.path.split(os.path.abspath(__file__))[0]
    with open(os.path.join(setup_dir, ".release_info")) as f:
        lines = []
        for line in f:
            lines.append(line)
    for x in lines[0].split(","):
        if "tag:" in x:
            version_nr = x.replace("tag:","").strip().lstrip("v")
        else:
            version_nr = ""
    revision_nr = lines[1].strip()[:7]
    date = lines[2].strip()
    # GitHub source archive (snapshot, no tag)
    if version_nr == "":
        with open(os.path.join(setup_dir, "CHANGES.md")) as f:
            for line in f:
                if line.lower().startswith("version"):
                    version_nr = "{0}-0-g{1}".format(line.split(" ")[1],
                                                    revision_nr)
                    break
    return version_nr, revision_nr, date

def get_version_info_from_file(filename):
    """Get version number, revision number and date from a .py file."""

    with open(filename) as f:
        for line in f:
            if line.startswith("__version__"):
                version_nr = line.split("'")[1]
            if line.startswith("__revision__"):
                revision_nr = line.split("'")[1]
            if line.startswith("__date__"):
                date = line.split("'")[1]
    return version_nr, revision_nr, date

def run():
    """Run the setup."""

    setup(name='expyriment',
          version=version_nr,
          description=description,
          author=author,
          author_email=author_email,
          license=license,
          url=url,
          packages=packages,
          package_dir=package_dir,
          package_data=package_data,
          data_files=data_files,
          install_requires=install_requires,
          extras_require=extras_require,
          cmdclass=cmdclass,
          entry_points = entry_points)

if __name__=="__main__":

    # Check if we are building/installing from a built archive/distribution
    version_nr, revision_nr, date = get_version_info_from_file("expyriment/__init__.py")
    if not version_nr == '':
        cmdclass={'install': Install,
                  'bdist_wininst': Wininst,
                  'install_data': InstallData,}
        run()
        message = "from built archive/distribution"

    # If not, we are building/installing from source
    else:
        cmdclass={'sdist': Sdist,
                  'build_py': Build,
                  'install': Install,
                  'bdist_wininst': Wininst,
                  'install_data': InstallData}

        # Are we building/installing from a source archive/distribution?
        version_nr, revision_nr, date = get_version_info_from_release_info()
        if not date.startswith("$Format:"):
            run()
            message = "from source archive/distribution"

        # Are we building/installing from the GitHub repository?
        else:
            try:
                proc = Popen(['git', 'rev-list', '--max-parents=0', 'HEAD'],
                             stdout=PIPE, stderr=PIPE)
                initial_revision = proc.stdout.readline()
                if not b'e21fa0b4c78d832f40cf1be1d725bebb2d1d8f10' in \
                                                                initial_revision:
                    raise Exception
                version_nr, revision_nr, date = get_version_info_from_git()
                run()
                message = "from repository"
            except:
                raise RuntimeError("Building/Installing Expyriment failed!")

    print("")
    print("Expyriment Version: [{0}] ({1})".format(version_nr, message))
    try:
        print("Warning:", warning)
    except:
        pass



"""
Input and output parallel port.

This module contains a class implementing parallel port input/output on Linux
and Windows.

The module is a modified version of the parallel module of PsychoPy
(www.psychopy.org).
The code in this file, as well as in the files '_inpout32.py', '_dlportio.py'
and '_linux.py' are heavily based on the original code.

Notes
-----
On Windows, the module will attempt to load whichever parallel port driver
(inpout32 or dlportio) is found first on the system.
On Linux, the PyParallel package is used.

Each instance of the class can provide access to a different parallel
port.

"""
from __future__ import absolute_import, print_function, division
from builtins import *

__author__ = 'Florian Krause <florian@expyriment.org> \
Oliver Lindemann <oliver@expyriment.org>'
__version__ = ''
__revision__ = ''
__date__ = ''


import sys
import os

from ... import _internals, misc
from ...io._input_output import Input, Output

if sys.platform.startswith('linux'):
    try:
        from ._linux import PParallelLinux
        _ParallelPort = PParallelLinux
        _ParallelPort._driver = "pyparallel"
    except:
        _ParallelPort = None
elif sys.platform == 'win32':
    try:
        from ._inpout32 import PParallelInpOut32
        _ParallelPort = PParallelInpOut32
        from ctypes import windll
        try:
            windll.inpout32
            _ParallelPort._driver = "inpout32"
        except:
            windll.inpoutx64
            _ParallelPort._driver = "inpoutx64"
    except:
        try:
            from ._dlportio import PParallelDLPortIO
            _ParallelPort = PParallelDLPortIO
            from ctypes import windll
            windll.dlportio
            _ParallelPort._driver = "dlportio"
        except:
            _ParallelPort = None
else: # MAC
    _ParallelPort = None
    # class PP(object):
    #     def __init__(self, address):
    #         self.address = address
    #     def read_data(self):
    #         return 255
    #     def read_status(self):
    #         return 32
    #     def read_control(self):
    #         return 16
    #     def set_control(self):
    #         pass
    #     def poll(self):
    #         return 8192
    #     def setData(self, data):
    #         pass
    #     def readData(self):
    #         return 100
    #     def readPin(self, pin):
    #         return False
    #     def setPin(self, pin, state):
    #         pass
    # _ParallelPort = PP


class ParallelPort(Input, Output):
    """A class implementing a parallel port input and output.

    Notes
    -----
    CAUTION: On Windows, one of the following parallel port drivers needs to
    be installed: 'inpout32' (http://www.highrez.co.uk/Downloads/InpOut32/) or
    'dlportio' (http://real.kiev.ua/2010/11/29/dlportio-and-32-bit-windows/).
    On Linux, the Python package 'PyParallel'
    (http://pyserial.sourceforge.net/pyparallel.html) has to be installed.

    """

    def __init__(self, address, reverse=False):
        """Create a parallel port input and output.

        Parameters
        ----------
        address : hex or str
            The address of the port to use.
        reverse : bool
            Whether the port should be in reverse mode (default=False).
            Reverse mode enables to read from the data pins.

        Notes
        -----
        On Windows, common port addresses are::

            LPT1 = 0x0378 or 0x03BC
            LPT2 = 0x0278 or 0x0378
            LPT3 = 0x0278

        On Linux, port addresses are in the following format::

            /dev/parport0

        """

        if _ParallelPort is None:
            if sys.platform == "win32":
                _message = "Please install one of the following parallel port " + \
"drivers: 'inpout32' (http://www.highrez.co.uk/Downloads/InpOut32/) or " + \
"'dlportio' (http://real.kiev.ua/2010/11/29/dlportio-and-32-bit-windows/)."
            elif sys.platform.startswith("linux"):
                _message = "Please install the Python package 'PyParallel'."
            else:
                _message = "Not available on your computer."
            message = "ParallelPort cannot be initialized! {0}".format(
                _message)
            raise ImportError(message)

        Input.__init__(self)
        Output.__init__(self)
        if isinstance(address, str) and address.startswith('0x'):
            address = int(address, 16)
        self._address = address
        try:
            self._parallel = _ParallelPort(address=address)
        except:
            raise RuntimeError(
                "Could not initiate parallel port at {0}".format(address))
        self.input_history = False  # dummy
        self._reverse = reverse

    @property
    def address(self):
        """Getter for address."""
        return self._address

    @property
    def parallel(self):
        """Getter for parallel."""
        return self._parallel

    @property
    def has_input_history(self):
        """Returns always False, because ParallelPort has no input history."""

        return False

    @property
    def reverse(self):
        """Getter for reverse."""

        return self._parallel.reverse

    @reverse.setter
    def reverse(self, value):
        """Setter for reverse."""

        self._parallel.reverse = value

    def clear(self):
        """Clear the parallel port.

        Dummy method required for port interfaces (see e.g. ButtonBox)

        """

        pass

    def read_data(self):
        """Read the data pins.

        Reads the data pins (2-9) on the parallel port.

        The received data is encoded in 8 bits, corresponding to the 8 pins::

            Pin:  9   7   6   5   4   3   2   1
                  |   |   |   |   |   |   |   |
            Bit: 128  64  32  16  8   4   2   1

        Returns
        -------
        data : int
            The value of the data pins.

        """

        data = self._parallel.readData()
        if self._logging:
            _internals.active_exp._event_file_log(
                    "ParallelPort,received,{0},read_data".format(data))
        return data

    def read_status(self):
        """Read data from the status pins.

        Reads all status pins (10, 11, 12, 13, 15).

        The received data is encoded in 5 bits, corresponding to the 5 pins::

            Pin: 11  10  12  13  15
                  |   |   |   |   |
            Bit: 16   8   4   2   1

        Returns
        -------
        data : int (0-31)
            The value of the status pins.

        """

        # TODO: Proper implementation that does not rely on single pins!

        bits = "{0}{1}{2}{3}{4}".format(int(self.read_pin(11)),
                                        int(self.read_pin(10)),
                                        int(self.read_pin(12)),
                                        int(self.read_pin(13)),
                                        int(self.read_pin(15)))
        data = int(bits, 2)
        if self._logging:
            _internals.active_exp._event_file_log(
                    "ParallelPort,received,{0},read_status".format(data), 2)
        return data

    def read_control(self):
        """Read data from the control pins.

        Reads control pins 1, 14, 16, 17.

        The received data is encoded in 4 bits, corresponding to the 4 pins::

            Pin: 17  16  14   1
                  |   |   |   |
            Bit:  8   4   2   1

        Returns
        -------
        data : int (0-15)
            The value of the control pins.

        """

        # TODO: Proper implementation that does not rely on single pins!

        bits = "{0}{1}{2}{3}".format(int(self.read_pin(17)),
                                     int(self.read_pin(16)),
                                     int(self.read_pin(14)),
                                     int(self.read_pin(1)))
        data = int(bits, 2)
        if self._logging:
            _internals.active_exp._event_file_log(
                    "ParallelPort,received,{0},read_control".format(data), 2)
        return data

    def poll(self):
        """Poll the parallel port.

        This will read the status pins (11, 10, 12, 13, 15), the data pins (2-9),
        and control pins (1, 14, 16, 17).

        The received data is encoded in 17 bits, corresponding to the 17 pins::

            Pin:  17    16    14     1     9     8     7     6     5
                   |     |     |     |     |     |     |     |     |
            Bit: 65536 32768 16384 8192  4096  2048  1024   512   256
             
            Pin:  4    3    2   11   10   12   13   15
                  |    |    |    |    |    |    |    |
            Bit: 128  64   32   16    8    4    2    1
             
        Returns
        -------
        data : int (0-65535)
            The value of all input pins.

        """

        c = self.read_control()
        d = self.read_data()
        s = self.read_status()

        data = (int('{:04b}'.format(c)[::1], 2) << 13) + \
               (int('{:08b}'.format(d)[::1], 2) << 5) + \
               s
        if self._logging:
            _internals.active_exp._event_file_log(
                    "ParallelPort,received,{0},poll".format(data), 2)
        return data

    def read_pin(self, pin):
        """Determine whether an input pin is set high(True) or low(False).

        Parameters
        ----------
        pin : int
            The number of the input pin to read (2-13 and 15)

        Returns
        -------
        set : bool
            Whether the pin is set high(True) or low(False)

        """

        return self._parallel.readPin(pin)

    def set_data(self, data):
        """Send data via data pins.

        Sets the data pins (2-9) on the parallel port.

        The data is encoded in 8 bits, corresponding to the 8 pins::

            Pin:  9   7   6   5   4   3   2   1
                  |   |   |   |   |   |   |   |
            Bit: 128  64  32  16  8   4   2   1

        Parameters
        ----------
        data : int (0-255)
            The data to be send via the data pins.

        """

        self._parallel.setData(data)
        if self._logging:
            _internals.active_exp._event_file_log(
                                    "ParallelPort,set_data,{0}".format(data), 2)

    def set_control(self, data):
        """Send data via control pins.

        Sets the control pins (1, 14, 16, 17) on the parallel port.

        The data is encoded in 8 bits, corresponding to the 4 pins::

            Pin: 17  16  14   1
                  |   |   |   |
            Bit:  8   4   2   1

        Parameters
        ----------
        data : int (0-15)
            The data to be send via the control pins.

        """

        # TODO: Proper implementation that does not rely on single pins!

        self._parallel.setPin(1, data & 1)
        self._parallel.setPin(14, data & 2)
        self._parallel.setPin(16, data & 4)
        self._parallel.setPin(17, data & 8)
        if self._logging:
            _internals.active_exp._event_file_log(
                                    "ParallelPort,set_control,{0}".format(data), 2)

    def send(self, data):
        """Send data via all output pins.

        Sets the data pins (2-9) and the control pins (1, 14, 16, 17) on the
        parallel port.

        The data is encoded in 12 bits, corresponding to the 12 pins::

            Pin:  17   16   14    1    9    8    7    6    5    4    3    2
                   |    |    |    |    |    |    |    |    |    |    |    |
            Bit: 2048 1024  512  256  128  64   32   16    8    4    2    1

        Parameters
        ----------
        data : int (0-2047)
            The data to be send via the output pins.

        """

        # TODO: Proper implementation that does not rely on single pins!

        d = data & 255
        c = data & 2048 >> 8
        self._parallel.setData(d)
        self._parallel.setPin(1, c & 1)
        self._parallel.setPin(14, c & 2)
        self._parallel.setPin(16, c & 4)
        self._parallel.setPin(17, c & 8)
        if self._logging:
            _internals.active_exp._event_file_log(
                                    "ParallelPort,send,{0}".format(data), 2)

    def set_pin(self, pin, state):
        """Set a desired output pin to be high(True) or low(False).

        Parameters
        ----------
        pin : int
            The number of the output pin to set (2-9).
        state : bool
            Whether the pin is to be set high(True) or low(False).

        """

        self._parallel.setPin(pin, int(state))

    @staticmethod
    def get_driver():
        """Return the driver used for interacting with parallel ports.
        
        On Windows, one of the following parallel port drivers needs to
        be installed: 'input32' (http://www.highrez.co.uk/Downloads/InpOut32/) or
        'dlportio' (http://real.kiev.ua/2010/11/29/dlportio-and-32-bit-windows/).
        On Linux, the Python package 'PyParallel'
        (http://pyserial.sourceforge.net/pyparallel.html) has to be installed.
    
        """
        
        if _ParallelPort is not None:
            return _ParallelPort._driver
        
    @staticmethod
    def get_available_ports():
        """Return an array of strings representing available parallel ports.

        Returns
        -------
        ports : list
            array of strings representing the available parallel ports

        Notes
        -----
        This method does only work on Linux.

        """

        ports = []
        if sys.platform.startswith("linux"):
            dev = os.listdir('/dev')
            for p in dev:
                if p.startswith("parport"):
                    ports.append(p)
        else:
            pass
        ports.sort()

        return ports

    @staticmethod
    def _self_test(exp):
        """Test the parallel port."""

        from ... import stimuli, io
        result = {}
        result["testsuite_parallel_port"] = ""
        result["testsuite_parallel_success"] = "No"

        if _ParallelPort is None:
            if sys.platform == "win32":
                _message = "Please install one of the following parallel port " + \
"drivers: 'inpout32' (http://www.highrez.co.uk/Downloads/InpOut32/) or " + \
"'dlportio' (http://real.kiev.ua/2010/11/29/dlportio-and-32-bit-windows/)."
            elif sys.platform.startswith("linux"):
                _message = "Please install the Python package 'PyParallel'."
            else:
                _message = "Not available on your computer."
            while True:
                stimuli.TextScreen(
                    "ParallelPort cannot be initialized!",
                    _message + "\n\n[Press RETURN to continue]").present()
                key, rt_ = exp.keyboard.wait(misc.constants.K_RETURN)
                if key is not None:
                    break
            return result

        else:
            text = "This will test if parallel port " \
                   "communication works correctly.\n\n" \
                   "You will be asked to enter a port address. " \
                   "On Windows this will be a hexadecimal number (e.g. " \
                   "'0x0378', which is ofter the address of LPT1). " \
                   "On Linux it will be of the form '/dev/parportX', "\
                   "where X denotes the number of the port.\n\n" \
                   "Once the port is opened, a list of control, status and " \
                   "data pins is presented. Bright green lights indicate " \
                   "input activity of a pin. Bright red lights indicate " \
                   "output activity of a pin. With the number keys on the " \
                   "keyboard, sending can be toggled on and off for each " \
                   "data pin. Reverse mode can be toggled on an of with " \
                   "the 'r' key.\n\n\n\n" \
                   "[Press RETURN to continue]"
            t = stimuli.TextScreen("ParallelPort test", text)
            while True:
                t.present()
                key, rt_ = exp.keyboard.wait(misc.constants.K_RETURN)
                if key is not None:
                    break

            address = io.TextInput("Port address:").get()
            try:
                pp = io.ParallelPort(address=address)
            except:
                while True:
                    stimuli.TextScreen(
                        "Could not open port address {0}!".format(address),
                        "[Press RETURN to continue]").present()
                    key, rt_ = exp.keyboard.wait(misc.constants.K_RETURN)
                    if key is not None:
                        break
                return result

            pin_canvas = stimuli.Canvas((800, 600))
            poll = stimuli.TextLine("Poll", text_colour=[255, 255, 255],
                                    position=[0, 165])
            poll.plot(pin_canvas)
            poll_line = stimuli.Rectangle((660, 2), colour=[255, 255, 255],
                                          position=[0, 120])
            poll_line.plot(pin_canvas)
            control = stimuli.TextLine("Control", text_colour=[100, 100, 100],
                                       position=[-260, 95])
            control.plot(pin_canvas)
            control_line = stimuli.Rectangle((140, 2), colour=[100, 100, 100],
                                             position=[-260, 50])
            control_line.plot(pin_canvas)
            data = stimuli.TextLine(
                "Data",
                text_colour=misc.constants.C_EXPYRIMENT_ORANGE,
                position=[-20, 95])
            data.plot(pin_canvas)
            data_line = stimuli.Rectangle(
                (300, 2),
                colour=misc.constants.C_EXPYRIMENT_ORANGE,
                position=[-20, 50])
            data_line.plot(pin_canvas)
            status = stimuli.TextLine(
                "Status",
                text_colour=misc.constants.C_EXPYRIMENT_PURPLE,
                position=[240, 95])
            status.plot(pin_canvas)
            status_line = stimuli.Rectangle(
                (180, 2),
                colour=misc.constants.C_EXPYRIMENT_PURPLE,
                position=[240, 50])
            status_line.plot(pin_canvas)

            pins = {}
            inputs = {}
            outputs = {}
            x_pos = -320
            for pin in [17, 16, 14, 1, 9, 8, 7, 6, 5, 4, 3, 2, 11, 10, 12, 13, 15]:
                inputs[pin] = stimuli.Circle(10, anti_aliasing=10,
                                             position=[x_pos, 30])
                if 1 < pin < 10:
                    colour = misc.constants.C_EXPYRIMENT_ORANGE
                elif pin in [10, 11, 12, 13, 15]:
                    colour = misc.constants.C_EXPYRIMENT_PURPLE
                else:
                    colour = [100, 100, 100]
                if pin in [11, 1, 14, 17]:
                    bg = stimuli.Rectangle((25, 20), colour=colour,
                                           position=[x_pos, 0])
                    bg.plot(pin_canvas)
                    colour = _internals.active_exp._background_colour
                stim = stimuli.TextLine(repr(pin), text_font="freemono",
                                        text_colour=colour,
                                        position=[x_pos, 0])
                stim.plot(pin_canvas)
                pins[pin] = stim
                if pin < 10 or pin in [14, 16, 17]:
                    outputs[pin] = stimuli.Rectangle(
                        (20, 20), position=[x_pos, -30])
                x_pos += 40

            control2 = stimuli.TextLine("Control", text_colour=[100, 100, 100],
                                        position=[-260, -95])
            control2.plot(pin_canvas)
            control2_line = stimuli.Rectangle((140, 2), colour=[100, 100, 100],
                                              position=[-260, -50])
            control2_line.plot(pin_canvas)

            data2 = stimuli.TextLine(
                "Data",
                text_colour=misc.constants.C_EXPYRIMENT_ORANGE,
                position=[-20, -95])
            data2.plot(pin_canvas)
            data2_line = stimuli.Rectangle(
                (300, 2),
                colour= misc.constants.C_EXPYRIMENT_ORANGE,
                position=[-20, -50])
            data2_line.plot(pin_canvas)

            send = stimuli.TextLine("Send", text_colour=[255, 255, 255],
                                    position=[-100, -165])
            send.plot(pin_canvas)
            send_line = stimuli.Rectangle((460, 2), colour=[255, 255, 255],
                                          position=[-100, -120])

            send_line.plot(pin_canvas)

            end = stimuli.TextLine(
                "[Press RETURN to end the test]",
                position=[0, -240])
            end.plot(pin_canvas)

            pin_canvas.present()
            pin_canvas.present()
            pin_canvas.present()

            def _update(inputs_states, outputs_states, read_control,
                        read_data, read_status, read_poll):
                for pin in list(inputs.keys()):
                    if inputs_states[pin] == True:
                        inputs[pin].colour = [0, 255, 0]
                    else:
                        inputs[pin].colour = [0, 30, 0]
                    inputs[pin].present(clear=False, update=False)
                    inputs[pin].present(clear=False, update=False)
                    inputs[pin].present(clear=False, update=False)

                send = 0
                for pin in list(outputs.keys()):
                    if outputs_states[pin][0] == True:
                        send += outputs_states[pin][1]
                        outputs[pin].colour = [255, 0, 0]
                    else:
                        outputs[pin].colour = [30, 0, 0]
                    outputs[pin].present(clear=False, update=False)
                    outputs[pin].present(clear=False, update=False)
                    outputs[pin].present(clear=False, update=False)

                results_canvas = stimuli.Canvas(
                    (600, 20),
                    colour=_internals.active_exp.background_colour,
                    position=[0, 70])

                rc = stimuli.TextLine(
                    "{0}".format(read_control),
                    text_colour=[0, 255, 0],
                    position=[-260, 0])
                rc.plot(results_canvas)

                rd = stimuli.TextLine(
                    "{0}".format(read_data),
                    text_colour=[0, 255, 0],
                    position=[-20, 0])
                rd.plot(results_canvas)

                rs = stimuli.TextLine(
                    "{0}".format(read_status),
                    text_colour=[0, 255, 0],
                    position=[240, 0])
                rs.plot(results_canvas)

                poll_canvas = stimuli.Canvas(
                    (100, 20),
                    colour=_internals.active_exp.background_colour,
                    position=[0, 140])

                pd = stimuli.TextLine(
                    "{0}".format(read_poll),
                    text_colour=[0, 255, 0])
                pd.plot(poll_canvas)

                poll_canvas.present(clear=False, update=False)
                poll_canvas.present(clear=False, update=False)
                poll_canvas.present(clear=False, update=False)

                results_canvas.present(clear=False, update=False)
                results_canvas.present(clear=False, update=False)
                results_canvas.present(clear=False, update=False)

                results_canvas2 = stimuli.Canvas(
                    (600, 20),
                    colour=_internals.active_exp.background_colour,
                    position=[0, -70])

                if send > 255:
                    c = (send ^ 255) >> 8
                else:
                    c = 0
                sc = stimuli.TextLine(
                    "{0}".format(c),
                    text_colour=[255, 0, 0],
                    position=[-260, 0])
                sc.plot(results_canvas2)

                sd = stimuli.TextLine(
                    "{0}".format(send & 255),
                    text_colour=[255, 0, 0],
                    position=[-20, 0])
                sd.plot(results_canvas2)

                send_canvas = stimuli.Canvas(
                    (100, 20),
                    colour=_internals.active_exp.background_colour,
                    position=[-100, -140])

                ss = stimuli.TextLine(
                    "{0}".format(send),
                    text_colour=[255, 0, 0])
                ss.plot(send_canvas)

                if pp.reverse is True:
                    r = stimuli.Rectangle(
                        (150, 30),
                        line_width=1,
                        colour=misc.constants.C_EXPYRIMENT_ORANGE,
                        position=[240, -95])
                    reverse = stimuli.TextLine(
                        "Reverse Mode",
                        text_colour=misc.constants.C_EXPYRIMENT_ORANGE,
                        background_colour=_internals.active_exp.background_colour)
                    reverse.plot(r)
                else:
                    r = stimuli.Rectangle(
                        (150, 30),
                        line_width=1,
                        colour=_internals.active_exp.background_colour,
                        position=[240, -95])
                    reverse = stimuli.TextLine(
                        "Reverse Mode",
                        text_colour=_internals.active_exp.background_colour,
                        background_colour=_internals.active_exp.background_colour)
                    reverse.plot(r)

                r.present(clear=False, update=False)
                r.present(clear=False, update=False)
                r.present(clear=False, update=False)

                send_canvas.present(clear=False, update=False)
                send_canvas.present(clear=False, update=False)
                send_canvas.present(clear=False, update=False)

                results_canvas2.present(clear=False, update=False)
                results_canvas2.present(clear=False, update=False)
                results_canvas2.present(clear=False, update=False)

                exp.screen.update()

            inputs_states = {1: False, 2:False, 3:False, 4:False, 5:False,
                             6:False, 7:False, 8:False, 9:False, 10:False,
                             11:False, 12:False, 13:False, 14:False, 15:False,
                             16:False, 17:False}

            outputs_states = {1:[False, 256], 2:[False, 1], 3:[False, 2],
                              4:[False, 4], 5:[False, 8], 6:[False, 16],
                              7:[False, 32], 8:[False, 64], 9:[False, 128],
                              14:[False, 512], 16:[False, 1024],
                              17:[False, 2048]}

            pp.send(0)
            initial_poll = pp.poll()

            while True:
                c = pp.read_control()
                d = pp.read_data()
                s = pp.read_status()
                p = pp.poll()

                if p != initial_poll:
                    result["testsuite_parallel_success"] = "Yes"

                for pin, bit in {15:1, 13:2, 12:4, 10:8, 11:16, 2:32, 3:64,
                                 4:128, 5:256, 6:512, 7:1024, 8:2048,
                                 9:4096, 1:8192, 14:16384, 16:32768,
                                 17:65536}.items():
                    if bit & p > 0:
                        inputs_states[pin] = True
                    else:
                        inputs_states[pin] = False

                keys = exp.keyboard.read_out_buffered_keys()
                if misc.constants.K_1 in keys or \
                                misc.constants.K_KP1 in keys:
                    if misc.constants.K_4 in keys or \
                                misc.constants.K_KP4 in keys:
                        outputs_states[14][0] =  not outputs_states[14][0]
                        pp.set_pin(14, outputs_states[14][0])
                    elif misc.constants.K_6 in keys or \
                                misc.constants.K_KP6 in keys:
                        outputs_states[16][0] =  not outputs_states[16][0]
                        pp.set_pin(16, outputs_states[16][0])
                    elif misc.constants.K_7 in keys or \
                                misc.constants.K_KP7 in keys:
                        outputs_states[17][0] =  not outputs_states[17][0]
                        pp.set_pin(17, outputs_states[17][0])
                    else:
                        outputs_states[1][0] =  not outputs_states[1][0]
                        pp.set_pin(1, outputs_states[1][0])
                elif misc.constants.K_2 in keys or \
                                misc.constants.K_KP2 in keys:
                    outputs_states[2][0] =  not outputs_states[2][0]
                    pp.set_pin(2, outputs_states[2][0])
                elif misc.constants.K_3 in keys or \
                                misc.constants.K_KP3 in keys:
                    outputs_states[3][0] = not outputs_states[3][0]
                    pp.set_pin(3, outputs_states[3][0])
                elif misc.constants.K_4 in keys or \
                                misc.constants.K_KP4 in keys:
                    outputs_states[4][0] = not outputs_states[4][0]
                    pp.set_pin(4, outputs_states[4][0])
                elif misc.constants.K_5 in keys or \
                                misc.constants.K_KP5 in keys:
                    outputs_states[5][0] = not outputs_states[5][0]
                    pp.set_pin(5, outputs_states[5][0])
                elif misc.constants.K_6 in keys or \
                                misc.constants.K_KP6 in keys:
                    outputs_states[6][0] = not outputs_states[6][0]
                    pp.set_pin(6, outputs_states[6][0])
                elif misc.constants.K_7 in keys or \
                                misc.constants.K_KP7 in keys:
                    outputs_states[7][0] = not outputs_states[7][0]
                    pp.set_pin(7, outputs_states[7][0])
                elif misc.constants.K_8 in keys or \
                                misc.constants.K_KP8 in keys:
                    outputs_states[8][0] = not outputs_states[8][0]
                    pp.set_pin(8, outputs_states[8][0])
                elif misc.constants.K_9 in keys or \
                                misc.constants.K_KP9 in keys:
                    outputs_states[9][0] = not outputs_states[9][0]
                    pp.set_pin(9, outputs_states[9][0])
                elif misc.constants.K_r in keys:
                    pp.reverse = not pp.reverse
                elif misc.constants.K_RETURN in keys:
                    result["testsuite_parallel_port"] = address
                    return result
                _update(inputs_states, outputs_states, c, d, s, p)



"""
This file is part of the geometry module.

This module contains functions to create vertex lists of basic
geometrical shapes.

"""

from __future__ import absolute_import, print_function, division
from builtins import *

__author__ = 'Florian Krause <florian@expyriment.org>, \
Oliver Lindemann <oliver@expyriment.org>'
__version__ = ''
__revision__ = ''
__date__ = ''

import math as _math
from ._geometry import XYPoint, points2vertices

def _angular_vertex(angle, length):
    """Helper function.

    Calculates the vertex coordinates of a line with a particular
    length and angle (relative to the horizontal).

    """

    angle = _math.radians(angle)
    return -1*_math.cos(angle)*float(length), -1*_math.sin(angle)*float(length)


def vertices_rectangle(size):
    """Returns a list of vertices describing a rectangle.

    Notes
    -----
    The resulting vertices can be plotted with the class
    stimuli.Shape(vertex_list=...).

    Parameters
    ----------
    size : (int, int)
        size (width, height) of the rectangle

    Returns
    -------
    vtx : list of vertices

    """

    return  [ (size[0]-1, 0),
              (0, -size[1]+1),
              (-size[0]+1, 0),
              (0, size[1]-1)]



def vertices_cross(size, line_width):
    """Returns a list of vertices describing a cross.

    Notes
    -----
    The resulting vertices can be plotted with the class
    stimuli.Shape(vertex_list=...).

    Parameters
    ----------
    size : (int, int)
        xy, length of the horizontal (x) and vertical (y) line
    line_width : int
        width of the lines

    Returns
    -------
    vtx : list of vertices

    See Also
    --------
    expyriment.stimuli.FixCross

    """

    x_a = (size[0] - line_width) // 2
    x_b = x_a
    y_a = (size[1] - line_width) // 2
    y_b = y_a

    if (size[0] - line_width) % 2:  # both have the different parities
        x_b = x_a + 1
        # to ensure that Shape behaves like two crossed surfaces plotted on each other
        if line_width % 2:  # odd line width swap x_a - x_b
            x_b, x_a = x_a, x_b

    if (size[1] - line_width) % 2:  # both have the different parities
        y_b = y_a + 1
        if line_width % 2 == 0:  # even line width swap x_a - x_b
            y_b, y_a = y_a, y_b

    return [(line_width - 1, 0),
            (0, -y_a),
            (x_a, 0),
            (0, -line_width + 1),
            (-x_a, 0),
            (0, -y_b),
            (-line_width + 1, 0),
            (0, y_b),
            (-x_b, 0),
            (0, line_width - 1),
            (x_b, 0)]


def vertices_trapezoid(width_top, width_bottom, height):
    """Returns a list of vertices describing a trapezoid

    Notes
    -----
    The resulting vertices can be plotted with the class
    stimuli.Shape(vertex_list=...).

    Parameters
    ----------
    width_top: int
        width of the top edge
    width_bottom: int
        width of the bottom edge
    height : int
        height of the trapezoid

    Returns
    -------
    vtx : list of vertices

    """

    left_bottom =XYPoint(x = 0, y = 0)
    right_bottom = XYPoint(x = width_bottom, y = 0)
    left_top = XYPoint(x = 0 + (width_bottom-width_top)/2.0, y = height)
    right_top = XYPoint(x = width_bottom - (width_bottom-width_top)/2.0, y = height)
    return list(map(lambda xy: (int(xy[0]), int(xy[1])),
                    points2vertices((left_top, right_top, right_bottom, left_bottom))))


def vertices_triangle(angle, length1, length2):
    """Returns a list of vertices describing a triangle A, B, C.

    ::

            A --- B
                 .
                .
               C

    Notes
    -----
    The resulting vertices can be plotted with the class
    stimuli.Shape(vertex_list=...).

    Parameters
    ----------
    angle : float
        the angle between the lines AB and BC in degrees
    length1 : float
        the length between AB
    length2 : float
        the length between BC

    Returns
    -------
    vtx : list of vertices

    """

    xy = _angular_vertex(angle, length2)
    return [(length1-1, 0), (int(xy[0]), int(xy[1]))]


def vertices_parallelogram(angle, length1, length2):
    """Returns a list of vertices describing a parallelogram A, B, C, D.

    ::

            A --- B
           .     .
          .     .
         D --- C

    Notes
    -----
    The resulting vertices can be plotted with the class
    stimuli.Shape(vertex_list=...).

    Parameters
    ----------
    angle : float
        the angle between the lines AB and BC in degrees
    length1 : float
        the length between AB
    length2 : float
        the length between BC

    Returns
    -------
    vtx : list of vertices

    """

    vtx = vertices_triangle(angle=angle, length1=length1, length2=length2)
    vtx.append((-length1+1, 0))
    return vtx


def vertices_regular_polygon(n_edges, length):
    """Returns a list of vertices describing a regular polygon.

    Notes
    -----
    The resulting vertices can be plotted with the class
    stimuli.Shape(vertex_list=...).

    Parameters
    ----------
    n_edges : int
        the number of edges
    length : float
        the length of one side of the polygon

    Returns
    -------
    vtx : list of vertices

    """

    sum_of_angle = (n_edges - 2) * 180.0
    angle = 180 - (sum_of_angle / n_edges)
    x = 180
    vtx = []
    for _ in range(n_edges - 1):
        v = _angular_vertex(x, length=length)
        vtx.append((int(v[0]), int(v[1])))
        x += angle
    return vtx


def vertices_frame(size, frame_thickness):
    """Returns a list of vertices describing a frame

    Notes
    -----
    The resulting vertices can be plotted with the class
    stimuli.Shape(vertex_list=...).

    Parameters
    ----------
    size : (int, int)
        size (width, height) of the rectangle
    frame_thickness : int
        the thickness of the frame

    Returns
    -------
    vtx : list of vertices

    """

    return  [(size[0] - frame_thickness - 1, 0),
             (0, -size[1]+1),
             (-size[0]+1, 0),
             (0, size[1]-1),
             (frame_thickness - 1, 0),
             (0, -(size[1] - frame_thickness - 1)),
             (size[0] - 2 * frame_thickness - 1, 0),
             (0, size[1] - 2 * frame_thickness - 1),
             (-(size[0] - 2 * frame_thickness - 2), 0)]



"""
The control._experiment_control module of expyriment.

"""
from __future__ import absolute_import, print_function, division
from builtins import *


__author__ = 'Florian Krause <florian@expyriment.org>, \
Oliver Lindemann <oliver@expyriment.org>'
__version__ = ''
__revision__ = ''
__date__ = ''

import sys
import os
import pygame
try:
    import android.mixer as mixer
except ImportError:
    import pygame.mixer as mixer

from . import defaults
from ._miscellaneous import _set_stdout_logging
from .._internals import get_version, android
from .. import design, stimuli, misc, _internals
from ..io import DataFile, EventFile, TextInput, Keyboard, Mouse, \
                _keyboard, TouchScreenButtonBox
from ..io._screen import Screen


def start(experiment=None, auto_create_subject_id=None, subject_id=None,
            skip_ready_screen=False):
    """Start an experiment.

    This starts an experiment defined by 'experiment' and asks for the subject
    number. When the subject number is entered and confirmed by ENTER, a data
    file is created.
    Eventually, "Ready" will be shown on the screen and the method waits for
    ENTER to be pressed.

    After experiment start the following additional properties are available:

    * experiment.subject -- the current subject id
    * experiment.data    -- the main data file

    Parameters
    ----------
    experiment : design.Experiment, optional (DEPRECATED)
        don't use this parameter, it only exists to keep backward compatibility
    auto_create_subject_id : bool, optional
        if True new subject id will be created automatically
    subject_id : integer, optional
        start with a specific subject_id;
        no subject id input mask will be presented; subject_id must be an
        integer; setting this parameter overrules auto_create_subject_id
    skip_ready_screen : bool, optional
        if True ready screen will be skipped (default=False)

    Returns
    -------
    exp : design.Experiment
        The started experiment will be returned.

    """

    if experiment is None:
        experiment = _internals.active_exp
    if experiment != _internals.active_exp:
        raise Exception("Experiment is not the currently initialized " +
                        "experiment!")
    if experiment.is_started:
        raise Exception("Experiment is already started!")
    if subject_id is not None:
        if not isinstance(subject_id, int):
            raise Exception("Subject id must be an integer. " +
                    "{0} is not allowed.".format(type(subject_id)))
        auto_create_subject_id = True
    elif auto_create_subject_id is None:
        auto_create_subject_id = defaults.auto_create_subject_id

    experiment._is_started = True
    # temporarily switch off log_level
    old_logging = experiment.log_level
    experiment.set_log_level(0)
    screen_colour = experiment.screen.colour
    experiment._screen.colour = [0, 0, 0]
    if subject_id is None:
        default_number = DataFile.get_next_subject_number()
    else:
        default_number = subject_id

    if not auto_create_subject_id:
        if android is not None:
            background_stimulus = stimuli.BlankScreen(colour=(0, 0, 0))
            fields = [stimuli.Circle(radius=100, colour=(70, 70, 70),
                                     position=(0, 70), anti_aliasing=10),
                      stimuli.Circle(radius=100, colour=(70, 70, 70),
                                     position=(0, -70), anti_aliasing=10),
                      stimuli.Rectangle(size=(50, 50), colour=(70, 70, 70),
                                        position=(120, 0))]
            fields[0].scale((0.25, 0.25))
            fields[1].scale((0.25, 0.25))
            plusminus = [
                stimuli.TextLine("Subject Number:", text_size=24,
                                 text_colour=misc.constants.C_EXPYRIMENT_PURPLE,
                                 position=(-182, 0)),
                stimuli.FixCross(size=(15, 15), position=(0, 70),
                                 colour=(0, 0, 0), line_width=2),
                stimuli.FixCross(size=(15, 2), position=(0, -70),
                                 colour=(0, 0, 0), line_width=2),
                stimuli.TextLine(text = "Go", text_size=18, position=(120, 0),
                                 text_colour=(0, 0, 0))]
            subject_id = default_number
            while True:
                text = stimuli.TextLine(
                    text="{0}".format(subject_id),
                    text_size=28,
                    text_colour=misc.constants.C_EXPYRIMENT_ORANGE)
                btn = TouchScreenButtonBox(
                    button_fields=fields,
                    stimuli=plusminus+[text],
                    background_stimulus=background_stimulus)
                btn.show()
                key, rt = btn.wait()
                if key == fields[0]:
                    subject_id += 1
                elif key == fields[1]:
                    subject_id -= 1
                    if subject_id <= 0:
                        subject_id = 0
                elif key == fields[2]:
                    break
            experiment._subject = int(subject_id)

        else:
            position = (0, 0)
            while True:
                ask_for_subject = TextInput(
                    message="Subject Number:",
                    position=position,
                    message_colour=misc.constants.C_EXPYRIMENT_PURPLE,
                    message_text_size=int(experiment.text_size * 1.2),
                    user_text_colour=misc.constants.C_EXPYRIMENT_ORANGE,
                    user_text_size=int(experiment.text_size),
                    background_colour=(0, 0, 0),
                    frame_colour=(70, 70, 70),
                    character_filter=misc.constants.K_ALL_DIGITS)
                subject_id = ask_for_subject.get(repr(default_number))
                try:
                    experiment._subject = int(subject_id)
                    break
                except:
                    pass

    else:
        experiment._subject = default_number

    experiment.screen.clear()
    experiment.screen.update()
    experiment._data = DataFile(additional_suffix=experiment.filename_suffix)
    experiment.data.add_variable_names(experiment.data_variable_names)
    for txt in experiment.experiment_info:
        experiment.data.add_experiment_info(txt)
    for line in experiment.__str__().splitlines():
        experiment.data.add_experiment_info(line)

    for f in experiment.bws_factor_names:
        _permuted_bws_factor_condition = \
            experiment.get_permuted_bws_factor_condition(f)
        experiment.data.add_subject_info(u"{0} = {1}".format(
            f, _permuted_bws_factor_condition))

    if experiment.events is not None:
        experiment.events._time_stamp = experiment.data._time_stamp
        experiment.events.rename(experiment.events.standard_file_name)

    number = defaults.initialize_delay - int(experiment.clock.time // 1000)
    if number > 0:
        text = stimuli.TextLine("Initializing, please wait...",
                                text_size=int(experiment.text_size * 1.2),
                                text_colour=(160, 70, 250),
                                position=(0, 0))
        stimuli._stimulus.Stimulus._id_counter -= 1
        text.present()
        text.present()  # for flipping with double buffer
        text.present()  # for flipping with tripple buffer
    default_textline_size = stimuli.TextLine(text="").text_size
    while number > 0:
        counter = stimuli.TextLine(
            "{num:02d}".format(num=number),
            text_size= int(experiment.text_size * 0.9),
            text_font='FreeMono',
            text_bold=True,
            text_colour=misc.constants.C_EXPYRIMENT_ORANGE,
            position=(0, -50),
            background_colour=(0, 0, 0))

        stimuli._stimulus.Stimulus._id_counter -= 1
        counter.present(clear=False)
        number -= 1
        key = experiment.keyboard.wait(pygame.K_ESCAPE, duration=1000,
                                       process_control_events=False)
        if key[0] is not None:
            break

    position = (0, 0)
    if not skip_ready_screen:
        stimuli.TextLine("Ready", position=position,
                         text_size=int(experiment.text_size * 1.2),
                         text_colour=misc.constants.C_EXPYRIMENT_ORANGE).present()
        stimuli._stimulus.Stimulus._id_counter -= 1
        if android is None:
            experiment.keyboard.wait()
        else:
            experiment.mouse.wait_press()
    experiment.set_log_level(old_logging)
    experiment._screen.colour = screen_colour
    experiment.log_design_to_event_file()
    experiment._event_file_log("Experiment,started")
    return experiment


def pause():
    """Pause a running experiment.

    This will show a pause screen and waits for ENTER to be pressed to
    continue.

    """

    if not _internals.active_exp.is_initialized:
        raise Exception("Experiment is not initialized!")
    experiment = _internals.active_exp
    experiment._event_file_log("Experiment,paused")
    screen_colour = experiment.screen.colour
    experiment._screen.colour = [0, 0, 0]
    old_logging = experiment.log_level
    experiment.set_log_level(0)
    if android is not None:
        position = (0, 200)
    else:
        position = (0, 0)
    stimuli.TextLine("Paused", position=position,
                     text_colour=misc.constants.C_EXPYRIMENT_ORANGE,
                     text_size=int(_internals.active_exp.text_size * 1.2)).present()
    experiment.set_log_level(old_logging)
    experiment._screen.colour = screen_colour
    stimuli._stimulus.Stimulus._id_counter -= 1
    misc.Clock().wait(200)
    if android is None:
        experiment.keyboard.wait()
    else:
        experiment.mouse.wait_press()
    experiment._event_file_log("Experiment,resumed")


def end(goodbye_text=None, goodbye_delay=None, confirmation=False,
        fast_quit=None, system_exit=False, pre_quit_function=None):
    """End expyriment.

    Parameters
    ----------
    goodbye_text  : str, optional
        text to present on the screen when quitting
    goodbye_delay : int, optional
        period to show the goodbye_text
    confirmation : bool, optional
        ask for confirmation (default = False)
    fast_quit : bool, optional
        quit faster by hiding the screen before actually quitting
        (default = None)
    system_exit : bool, optional
        call Python's sys.exit() method when ending expyriment (default = False)
    pre_quit_function : function, optional
        function to be called before quitting Pygame

    Returns
    -------
    out : bool
        True if Expyriment (incl. Pygame) has been quit.

    """

    if not _internals.active_exp.is_initialized:
        if pre_quit_function is not None:
            pre_quit_function()

        # Delete open file handles and previously opened fonts
        import expyriment.stimuli._textline, expyriment.stimuli._textbox
        for text_stim in [stimuli._textline, stimuli._textbox]:
            for f in text_stim.open_filehandles:
                f.close()
            text_stim.fonts = {}

        pygame.quit()
        if system_exit:
            sys.exit()
        return True

    experiment = _internals.active_exp
    if confirmation:
        experiment._event_file_log("Experiment,paused")
        screen_colour = experiment.screen.colour
        experiment._screen.colour = [0, 0, 0]
        if android is not None:
            position = (0, 200)
        else:
            position = (0, 0)
        stimuli.TextLine("Quitting Experiment? (y/n)", position=position,
                         text_colour=misc.constants.C_EXPYRIMENT_ORANGE,
                         text_size=int(_internals.active_exp.text_size * 1.2)).present()
        stimuli._stimulus.Stimulus._id_counter -= 1
        char = Keyboard().wait_char(["y", "n"], process_control_events=False)
        if char[0] == "n":
            experiment._screen.colour = screen_colour
            experiment._event_file_log("Experiment,resumed")
            return False
    experiment._event_file_log("Experiment,ended")
    if goodbye_text is None:
        goodbye_text = defaults.goodbye_text
    if goodbye_delay is None:
        goodbye_delay = defaults.goodbye_delay
    if experiment.events is not None:
        experiment.events.save()
    if experiment.data is not None:
        experiment.data.save()
    if fast_quit is None:
        fast_quit = defaults.fast_quit
    if fast_quit and experiment.is_started:
        if experiment.screen.window_mode:
            if experiment.screen.no_frame:
                pygame.display.set_mode(experiment.screen._window_size, pygame.NOFRAME)
            else:
                pygame.display.set_mode(experiment.screen._window_size)
            pygame.display.iconify()

    try:
        experiment._screen.colour = [0, 0, 0]
        stimuli.TextLine(goodbye_text, position=(0, 0),
                         text_colour=misc.constants.C_EXPYRIMENT_PURPLE,
                         text_size=int(_internals.active_exp.text_size * 1.2)).present()
        stimuli._stimulus.Stimulus._id_counter -= 1
    except:
        pass

    if not fast_quit:
        misc.Clock().wait(goodbye_delay)
    _internals.active_exp = design.Experiment("None")
    if pre_quit_function is not None:
        pre_quit_function()

    # Delete open file handles and previously opened fonts
    import expyriment.stimuli._textline, expyriment.stimuli._textbox
    for text_stim in [expyriment.stimuli._textline, expyriment.stimuli._textbox]:
        for f in text_stim.open_filehandles:
            f.close()
        text_stim.fonts = {}

    pygame.quit()
    if system_exit:
        sys.exit()
    return True


def initialize(experiment=None):
    """Initialize an experiment.

    This initializes an experiment defined by 'experiment' as well as the
    underlying expyriment system. If 'experiment' is None, a new Experiment
    object will be created and returned. Furthermore, a screen, a clock, a
    keyboard and a event file are created and added to the experiment. The
    initialization screen is shown for a short delay to ensure that Python
    is fully initialized and time accurate. Afterwards, "Preparing
    experiment..." is presented on the screen.

    After experiment initialize the following additional properties are
    available:

    - experiment.screen   -- the current screen
    - experiment.clock    -- the main clock
    - experiment.keyboard -- the main keyboard
    - experiment.mouse    -- the main mouse
    - experiment.event    -- the main event file

    Parameters
    ----------
    experiment : design.Experiment, optional
        the experiment to initialize

    Returns
    -------
    exp : design.Experiment
        initialized experiment

    """

    if experiment is None:
        experiment = design.Experiment()

    if experiment.log_level is None:
        experiment.set_log_level(defaults.event_logging)

    if misc.is_interactive_mode() and not defaults.window_mode \
        and not hasattr(experiment, "testsuite"):
        print("""
Python is running in an interactive shell but Expyriment wants to initialize a
fullscreen.""")
        quest = "Do you want to switch to windows mode?"
        ans = input(quest + " (Y/n) ").strip().lower()
        if ans=="" or ans=="y" or ans=="yes":
            print("Switched to windows mode")
            defaults.window_mode = True

    stdout_logging = defaults.stdout_logging
    _internals.active_exp = experiment
    old_logging = experiment.log_level
    experiment.set_log_level(0)  # switch off for the first screens

    _keyboard.quit_key = defaults.quit_key
    _keyboard.pause_key = defaults.pause_key
    _keyboard.end_function = end
    _keyboard.pause_function = pause

    mixer.pre_init(defaults.audiosystem_sample_rate,
                   defaults.audiosystem_bit_depth,
                   defaults.audiosystem_channels,
                   defaults.audiosystem_buffer_size)
    if defaults.audiosystem_autostart:
        mixer.init()
        mixer.init()  # Needed on some systems

    experiment._clock = misc.Clock()
    experiment._screen = Screen(colour=(0, 0, 0),
                                open_gl=defaults.open_gl,
                                window_mode=defaults.window_mode,
                                window_size=defaults.window_size,
                                no_frame=defaults.window_no_frame)
    # Hack for IDLE: quit pygame and call atexit functions when crashing
    if misc.is_idle_running() and sys.argv[0] != "":
        try:
            import idlelib.run

            def wrap(orig_func):
                def newfunc(*a, **kw):
                    pygame.quit()
                    import atexit
                    atexit._run_exitfuncs()
                    idlelib.run.flush_stdout = orig_func
                    return orig_func(*a, **kw)
                return newfunc
            idlelib.run.flush_stdout = wrap(idlelib.run.flush_stdout)
        except ImportError:
            pass
    experiment._data = None
    experiment._subject = None
    experiment._is_initialized = True  # required before EventFile
    if old_logging> 0:
        experiment._events = EventFile(
            additional_suffix=experiment.filename_suffix, time_stamp=True)
        if stdout_logging:
            _set_stdout_logging(experiment._events)
    else:
        experiment._events = None
    experiment._keyboard = Keyboard()
    experiment._mouse = Mouse(show_cursor=False)
    logo = stimuli.Picture(misc.constants.EXPYRIMENT_LOGO_FILE,
                           position=(0, 100))
    logo.scale((0.7, 0.7))
    text = stimuli.TextLine("Version {0}".format(get_version()),
                            text_size= experiment.text_size,
                            text_colour=misc.constants.C_EXPYRIMENT_PURPLE,
                            background_colour=(0, 0, 0),
                            position=(0, -5))
    canvas = stimuli.Canvas((600, 400), colour=(0, 0, 0))
    canvas2 = stimuli.Canvas((600, 400), colour=(0, 0, 0))
    logo.plot(canvas)
    text.plot(canvas)
    hash_ = misc.get_experiment_secure_hash()
    if hash_ is not None:
        txt = "{0} ({1})".format(os.path.split(sys.argv[0])[1], hash_)
        if len(misc.module_hashes_as_string())>0:
            txt += ", {0}".format(
                        misc.module_hashes_as_string())
        text2 = stimuli.TextLine(txt,
            text_size= int(experiment.text_size * 0.7),
            text_colour=misc.constants.C_EXPYRIMENT_ORANGE,
            background_colour=(0, 0, 0),
            position=(0, -50))
        text2.plot(canvas)
    canvas.preload(True)
    canvas._set_surface(canvas._get_surface().convert())
    start = experiment.clock.time
    r = [x for x in range(256) if x % 5 == 0]
    stopped = False
    if defaults.initialize_delay > 0:
        for x in r:
            canvas._get_surface().set_alpha(x)
            canvas2.clear_surface()
            canvas.plot(canvas2)
            canvas2.present()
            experiment.clock.wait(1)
            key = experiment.keyboard.check(pygame.K_ESCAPE,
                                            check_for_control_keys=False)
            if key is not None:
                stopped = True
                break
        duration = experiment.clock.time - start
        if duration < 2000 and not stopped:
            start = experiment.clock.time
            while experiment.clock.time - start < 2000:
                key = experiment.keyboard.check(pygame.K_ESCAPE,
                                                check_for_control_keys=False)
                if key is not None:
                    stopped = True
                    break
        r = [x for x in range(256)[::-1] if x % 5 == 0]
        if not stopped:
            for x in r:
                canvas._get_surface().set_alpha(x)
                canvas2.clear_surface()
                canvas.plot(canvas2)
                canvas2.present()
                experiment.clock.wait(1)
                key = experiment.keyboard.check(pygame.K_ESCAPE,
                                                check_for_control_keys=False)
                if key is not None:
                    break
    stimuli.TextLine("Preparing experiment...", text_size=int(experiment.text_size * 1.2),
                     text_colour=misc.constants.C_EXPYRIMENT_PURPLE).present()
    experiment._screen.colour = experiment.background_colour
    experiment.set_log_level(old_logging)
    stimuli._stimulus.Stimulus._id_counter = 0
    return experiment



#!/usr/bin/env python

"""
The tone stimulus module.

This module contains a class implementing a tone stimulus.

"""
from __future__ import absolute_import, print_function, division
from builtins import *


__author__ = 'Florian Krause <florian@expyriment.org>, \
Oliver Lindemann <oliver@expyriment.org>'
__version__ = ''
__revision__ = ''
__date__ = ''


import os
import math
import wave
import struct
import itertools
import tempfile
import shutil

from . import defaults
from ._audio import Audio


class Tone(Audio):
    """A class implementing a tone stimulus."""

    _getter_exception_message = "Cannot set {0} if preloaded!"

    def __init__(self, duration, frequency=None, samplerate=None,
                 bitdepth=None, amplitude=None):
        """Create a Tone.

        Parameters
        ----------
        duration : str
            duration of the file in ms
        frequency : int, optional
            frequency of the sine tone
        samplerate : int, optional
            samplerate of the sine tone
        bitdepth : int, optional
            bitdeth of the sine tone
        amplitude : int, optional
            amplitude of the sine tone

        """

        self._duration = duration / 1000.0
        if frequency is None:
            frequency = defaults.tone_frequency
        self._frequency = frequency
        if samplerate is None:
            samplerate = defaults.tone_samplerate
        self._samplerate = samplerate
        if bitdepth is None:
            bitdepth = defaults.tone_bitdepth
        self._bitdepth = bitdepth
        if amplitude is None:
            amplitude = defaults.tone_amplitude
        self._amplitude = amplitude
        filename = self._create_sine_wave()
        Audio.__init__(self, filename)


    @property
    def duration(self):
        """Getter for duration."""

        return self._duration * 1000.0

    @duration.setter
    def duration(self, value):
        """Setter for duration."""

        if self.is_preloaded:
            raise AttributeError(Audio._getter_exception_message.format(
                "duration"))
        else:
            self._duration = value / 1000.0
            self._filename = self._create_sine_wave()

    @property
    def frequency(self):
        """Getter for frequency."""

        return self._frequency

    @frequency.setter
    def frequency(self, value):
        """Setter for frequency."""

        if self.is_preloaded:
            raise AttributeError(Audio._getter_exception_message.format(
                "frequency"))
        else:
            self._frequency = value
            self._filename = self._create_sine_wave()

    @property
    def samplerate(self):
        """Getter for samplerate."""

        return self._samplerate

    @samplerate.setter
    def samplerate(self, value):
        """Setter for samplerate."""

        if self.is_preloaded:
            raise AttributeError(Audio._getter_exception_message.format(
                "samplerate"))
        else:
            self._samplerate = value
            self._filename = self._create_sine_wave()

    @property
    def bitdepth(self):
        """Getter for bitdepth."""

        return self._bitdepth

    @bitdepth.setter
    def bitdepth(self, value):
        """Setter for bitdepth."""

        if self.is_preloaded:
            raise AttributeError(Audio._getter_exception_message.format(
                "bitdepth"))
        else:
            self._bitdepth = value
            self._filename = self._create_sine_wave()

    @property
    def amplitude(self):
        """Getter for amplitude."""

        return self._amplitude

    @amplitude.setter
    def amplitude(self, value):
        """Setter for amplitude."""

        if self.is_preloaded:
            raise AttributeError(Audio._getter_exception_message.format(
                "amplitude"))
        else:
            self._amplitude = value
            self._filename = self._create_sine_wave()

    def _grouper(self, n, iterable, fillvalue=None):
        """Write in chunks."""

        args = [iter(iterable)] * n
        return itertools.zip_longest(fillvalue=fillvalue, *args)

    def _create_sine_wave(self):
        """Create the sine wave."""

        period = int(self._samplerate / self._frequency)
        lookup_table = [float(self._amplitude) * \
                        math.sin(2.0 * math.pi * float(self._frequency) * \
                                 (float(i % period) / float(self._samplerate))) \
                        for i in range(period)]
        sine = (lookup_table[i % period] for i in itertools.count(0))
        channels = ((sine,),)
        n_samples = int(self._duration * self._samplerate)
        samples = itertools.islice(zip(
            *(map(sum, zip(*channel)) \
              for channel in channels)), n_samples)
        fid, filename = tempfile.mkstemp(dir=defaults.tempdir,
                        prefix="freq{0}_dur{1}_".format(self.frequency,
                                                       self.duration),
                        suffix=".wav")
        os.close(fid)
        w = wave.open(filename, 'w')
        w.setparams((1, self._bitdepth // 8, self._samplerate, n_samples, 'NONE',
                     'not compressed'))
        max_amplitude = float(int((2 ** (self._bitdepth)) / 2) - 1)
        for chunk in self._grouper(2048, samples):
            frames = b''.join(b''.join(struct.pack(
                b'h', int(max_amplitude * sample)) for sample in channels) \
                for channels in chunk if channels is not None)
            w.writeframesraw(frames)
        w.close()
        return filename

    def save(self, filename):
        """Save the sine tone to a file.

        Parameters
        ----------
        filename : str
            filename the sine tone should be saved to (str)

        """

        shutil.copy(self._filename, filename)


    @staticmethod
    def _demo(exp=None):
        if exp is None:
            from .. import control
            control.set_develop_mode(True)
            control.defaults.event_logging = 0
            control.start_audiosystem()
            exp_ = control.initialize()
        sine = Tone(duration=1000)
        sine.present()
        if exp is None:
            exp_.clock.wait(1000)



#!/usr/bin/env python
"""Make yaml files for OpenSesame help menu"""

from __future__ import absolute_import, print_function, division

__author__ = 'Florian Krause <florian@expyriment.org> \
Oliver Lindemann <oliver@expyriment.org>'
__version__ = ''

import os
import sys

from create_rst_api_reference import inspect_members, exclude, expyriment

T = "  "
HOME = "Official Website: /"
DOC = " /docs/{0}/"
FULLDOC = "Full Documentation:" + DOC
SUPPORT = """Community/Support:
  Newsletter: /newsletter
  Forum: /forum
  Mailinglist: /mailinglist
  IRC/Matrix chat: /chat
  Facebook: /facebook
  Twitter: /twitter
  Reddit: /reddit
  Google+: /googleplus
"""
YAML_FILE = "sitemap.yml"


def get_version():
    try:
        info = expyriment.misc.get_system_info()
    except: # expyriment < 0.9
        info = expyriment.get_system_info()
    ver = info['python_expyriment_version']
    if len(ver)>0:
        print("Expyriment version:", ver)
        return ver
    else:
        p = os.path.abspath(os.path.join(os.path.split(sys.argv[0])[0], '..', '..'))
        changes_md = os.path.join(p, "CHANGES.md")
        with open(changes_md, 'r') as fl:
            for line in fl:
                if line.startswith("Version"):
                    ver = line.split(" ")[1]
                    print("Expyriment version from CHANGES.md:", ver)
                    return ver

def parse_module(mod_name, tab, doc_path):
    modules, classes, methods, functions, attributes = inspect_members(mod_name)
    rtn  = []
    submodels_add_later = []
    if len(modules)>0:
        for m in modules:
            if m[0] not in exclude:
                if m[0]=="constants":
                    rtn.append(tab + m[0] + ":" + doc_path + \
                                    mod_name + "." + m[0] + ".html")
                elif m[0] == "extras":
                    submodels_add_later.append(m[0])
                elif m[0] != "defaults":
                    tmp = parse_module(mod_name=mod_name + "." + m[0],\
                                tab=tab+T, doc_path=doc_path)
                    if len(tmp)>0:
                        rtn.append(tab + m[0] + ":")
                        rtn.extend(tmp)

    if len(classes)>0:
        for cl in classes:
            if cl[0] not in exclude:
                rtn.append(tab + cl[0] + ":")
                rtn.append(tab + T + "__init__:" + doc_path + \
                        mod_name + ".html#" + mod_name + "." + cl[0])

                rtn.extend(parse_module(mod_name=mod_name + "." + cl[0], \
                            tab=tab+ T, doc_path=doc_path))

    if len(methods)>0:
        for m in methods:
            if m[0] not in exclude:
                rtn.append(tab + m[0] + ":" + doc_path + \
                            mod_name + ".html#" + mod_name + "." + m[0])

    if len(functions)>0:
        for func in functions:
            if func[0] not in exclude:
                rtn.append(tab + func[0] + ":" + doc_path + \
                                mod_name + ".html#" + mod_name + "."+ func[0])

    for m in submodels_add_later:
        tmp = parse_module(mod_name=mod_name + "." + m, \
                            tab=tab+T, doc_path=doc_path)
        if len(tmp)>0:
            rtn.append(tab + m + ":")
            rtn.extend(tmp)

    return rtn


def api_ref_structure(version):
    """api reference structure with text as yaml text"""
    rtn = []
    rtn.append("API Reference:")
    rtn.extend(parse_module(mod_name="expyriment", tab=T,
                            doc_path = DOC.format(version)))

    return "\n".join(rtn)

if __name__ == "__main__":

    print("creating sitemap:", YAML_FILE)
    version = get_version()
    with open(YAML_FILE, "w+") as fl:
        fl.write(HOME + "\n")
        fl.write(api_ref_structure(version) + "\n")
        fl.write(FULLDOC.format(version) + "\n")
        fl.write(SUPPORT+ "\n")



"""
Default settings for the control package. ::

    audiosystem_autostart : bool
        start the audiosystem when Expyriment is initialized

    audiosystem_bit_depth : int
        the audio bit depth; negative values mean signed sample values

    audiosystem_buffer_size: int
        the audio buffer size in samples

    audiosystem_channels : int
        the number of audio channels

    audiosystem_sample_rate : int
        the audio sample rate

    auto_create_subject_id : bool
        create a subject id automatically when starting an experiment

    event_logging : int or bool
        O/False = no event logging
        1       = normal event logging (logging of all input & output events)
        2       = intensive logging; logs much more; use only for debugging
                  purposes

        NOTES
        =====
        In most cases, it should be avoided to switch of logging (loglevel=0).
        If log files become to big due to certain repetitive events, it is
        suggested to switch of the logging of individual stimuli or IO event
        (see the method .set_logging() of this object).
        The logging of all events can also be changed via
        expyriment.design.Experiment().set_logging.

    fast_quit : bool
        quit immediately without showing a goodbye message

    goodbye_delay : int
        duration (in milliseconds) to show the goodbye text

    goodbye_text : str
        the text to be shown when ending an experiment

    initialize_delay : int
        a delay (in seconds) to wait before running an experiment to give
        Python time to start properly

        NOTES
        =====
        Python needs about 10 seconds to start up properly. Before that,
        timing accuracy can be worse.

    open_gl : int or bool
        0/False = no OpenGL (no vsync / no blocking)
        1       = OpenGL (vsync / no blocking)
        2       = OpenGL (vsync / blocking)
        3       = OpenGL (vsync / alternative blocking)

        NOTES
        =====
        no OpenGL (no vsync / no blocking)
        ----------------------------------
        * Short preloading times
            Pygame is used to write directly to the display
        * Unsynchronized display updating
            Pixels are starting to change at wherever the vertical retrace is
            at the moment of presentation

        This mode can be useful for dynamic displays, where an accurate
        stimulus timing is not required, but rather very fast display changes
        have high priority and stimuli have to be created and on the fly (e.g.
        questionnairs, video game-like scenarios).

        OpenGL (vsync / no blocking)
        ----------------------------
        * Longer preloading times
            OpenGL is used to write to the display and a conversion to an
            OpenGL texture is needed
        * Synchronized display updating
            Pixels are starting to change always when the vertical retrace is
            at the left top corner
        * Innaccurate presentation time reporting
            Exact time of pixels changing on screen is not known

        This mode can be useful for dynamic displays, where an accurate
        stimulus timing is not required, but rather fast and fluid (e.g.
        non-tearing) display changes have high priority and preloaded stimuli
        can be used (e.g. well-controlled motion stimuli).

        OpenGL (vsync / blocking)
        -------------------------
        * Longer preloading times
            OpenGL is used to write to the display and a conversion to an
            OpenGL texture is needed
        * Synchronized display updating
            Pixels are starting to change always when the vertical retrace is
            at the left top corner
        * Accurate presentation time reporting
            present() method will only return when all drawing has been
            finished

        This can be useful for static stimulus presentation, where getting an
        accurate  presentation time report is of highest priority. Since this
        will be the case for most psychological and neuroscientific settings,
        it is the default mode used by Expyriment.

        OpenGL (vsync / alternative blocking)
        -------------------------------------
        * Longer preloading times
            OpenGL is used to write to the display and a conversion to an
            OpenGL texture is needed
        * Synchronized display updating
            Pixels are starting to change always when the vertical retrace is
            at the left top corner
        * Accurate presentation time reporting (alternative method)
            present() method will only return when all drawing has been
            finished and the then empty back buffer is available for drawing
            on it again

        This can be useful for static stimulus presentation, where getting an
        accurate presentation time report is of highest priority and the
        former mode does not work correctly (due to different video card
        driver implementations).

    pause_key : int
        the key to be used for pausing an experiment

    quit_key : int
        the key to be used for ending an experiment

    refresh_key : int
        the key to be used for refreshing the screen

    stdout_logging: bool
        log errors and warnings which go to the standard output (terminal
        window)

    window_mode : bool
        run experiment in window instead of fullscreen

    window_size : (int, int)
        the size of the window when running in window mode

    window_no_frame : bool
        if True windows in window mode will be displayed without a frame;
        this parameter does not affect fullscreen mode

"""

__author__ = 'Florian Krause <florian@expyriment.org, \
Oliver Lindemann <oliver@expyriment.org>'
__version__ = ''
__revision__ = ''
__date__ = ''


from ..misc import constants as _constants


initialize_delay = 10  # After approximately 10 seconds Python is timecritical
auto_create_subject_id = False
goodbye_text = "Ending experiment..."
goodbye_delay = 3000
fast_quit = False
quit_key = _constants.K_ESCAPE
pause_key = None  # TODO for 1.0: Pause events should be removed

open_gl = 2
window_mode = False
window_size = (800, 600)
window_no_frame = False

event_logging = 1  # 1 = default, 2 = extensive, 0 or False = off
stdout_logging = True

audiosystem_autostart = True
audiosystem_sample_rate = 44100
audiosystem_bit_depth = -16  # Negative values mean signed sample values
audiosystem_channels = 2
audiosystem_buffer_size = 2048

_mode_settings = None




"""The expyriment randomize module.

This module contains various functions for randomizing data

"""
from __future__ import absolute_import, print_function, division
from builtins import *

__author__ = 'Florian Krause <florian@expyriment.org>,\
              Oliver Lindemann <oliver@expyriment.org>'
__version__ = ''
__revision__ = ''
__date__ = ''


from copy import copy as _copy
import random as _random

_random.seed()


def rand_int_sequence(first_elem, last_elem):
    """Return a randomised sequence of integers in given range.

    Parameters
    ----------
    first_elem : int
        first element of the range
    last_elem : int
        last element of the range

    Returns
    -------
    rnd_seq : list
        randomised sequence of integers in given range

    """

    list_ = list(range(first_elem, last_elem + 1))
    _random.shuffle(list_)
    return list_


def rand_int(a, b):
    """Return random integer in given range.

    Parameters
    ----------
    a : int
        first element of range
    b : int
        last element of range

    Returns
    -------
    rnd : int

    """

    return _random.randint(a, b)


def rand_element(list_):
    """Return a random element from a list

    Parameters
    ----------
    list_ : list

    Returns
    -------
    elem : a random element from the list

    """

    list_ = list(list_)
    return list_[_random.randint(0, len(list_) - 1)]


def coin_flip(head_bias=0.5):
    """Return randomly True (head) or False (tail).

    Parameters
    ----------
    head_bias : numeric, optional
        bias in favor of head (default=0.5, fair coin)

    Returns
    -------
    rnd : bool

    """

    if head_bias < 0 or head_bias > 1:
        raise RuntimeError("Head bias must be between 0 and 1!")

    return _random.random() <= head_bias


def rand_norm(a, b, mu=None, sigma=None):
    """Normally distributed random number in given range.

    Parameters
    ----------
    a : numeric
        lowest number in range
    b : numeric
        highest number in range
    mu : numeric, optional
        distribution mean, default: mid point of the interval [a, b]
    sigma : numeric, optional
        distribution standard deviation, default: (b-a)/6.0

    Returns
    -------
    rnd : numeric

    """

    if mu is None:
        mu = a + (b-a) / 2.0
    if sigma is None:
        sigma = (b-a) / 6.0

    r = _random.normalvariate(mu=mu, sigma=sigma)
    if r < a or r > b:
        return rand_norm(a=a, b=b)

    return r


def _compare_items(a, b):
    """Helper function for `shuffle_list` to compare two elements of a list"""
    from .._structure import Trial, Block # needs to be imported here because of circular dependency
    if (isinstance(a, Trial) and isinstance(b, Trial)) or\
       (isinstance(a, Block) and isinstance(b, Block)):
        return a.compare(b)
    else:
        return a == b


def shuffle_list(list_, max_repetitions=-1, n_segments=0):
    """Shuffle any list of objects. In place randomization of the list.

    Parameters
    ----------
    list_ : list
        the list to shuffle; if not a list, TypeError is raised
    max_repetitions : int, optional
        maximum number of allowed repetitions of one identical items; if no
        solution can be found (i.e., Python's recursion limit is reached), the
        function returns `False` and the list will be randomized without
        constrains (see Notes); default = -1
    n_segments : int, optional
        randomize list per segment, i.e., list will be divided into n equal
        sized segments and the order of elements within each segment will be
        randomized; if n_segments is < 2, this parameter has no effect;
        default = 0

    Returns
    -------
    success : bool
        returns if randomization was successful and fulfilled the specified
        constrains (see max_repetitions)

    Notes
    -----
    When shuffling lists of trials or blocks, IDs and added stimuli will be
    ignored to determine repetitions, because trial or block comparisons are
    based on the `compare`method (see documentation of `Trial` or `Block`).

    """

    if not isinstance(list_, list):
        raise TypeError("The parameter 'list_' is a {0}, but has to be list. ".format(type(list_).__name__))

    if n_segments is None:
        n_segments = 0
    if max_repetitions is None:
        max_repetitions = -1

    if n_segments > 1:
        l = 1 + (len(list_) - 1) // int(n_segments)
        for x in range(n_segments):
            t = (x + 1) * l
            if t > len(list_):
                t = len(list_)
            a = list_[l * x:t]
            _random.shuffle(a)
            list_[l * x:t] = a
    else:
        _random.shuffle(list_)

    if max_repetitions >= 0:
        # check constrains and remix
        reps = 0
        for i in range(len(list_)-1):
            if _compare_items(list_[i], list_[i+1]):
                reps += 1
                if reps > max_repetitions:
                    try:
                        return shuffle_list(list_=list_,
                                max_repetitions=max_repetitions,
                                n_segments=n_segments)
                    except:  # maximum recursion depth reached
                        return False
            else:
                reps = 0
    return True


def make_multiplied_shuffled_list(list_, xtimes):
    """Return the multiplied and shuffled (sectionwise) list.

    The function manifolds the list 'x times' and shuffles each
    and concatenates to the return new lists.

    Parameters
    ----------
    list_ : list
        list to be shuffled
    xtimes : int
        how often the list will be multiplied. If xtimes==0, an
        empty list will be returned.

    """

    newlist = []
    tmp = _copy(list(list_))
    for _i in range(0, xtimes):
        _random.shuffle(tmp)
        newlist.extend(tmp)
    return newlist




"""
The base module of io.

This module contains the base classes for input and output.

All classes in this module should be called directly via expyriment.io.*.

"""

from __future__ import absolute_import, print_function, division

__author__ = 'Florian Krause <florian@expyriment.org>, \
Oliver Lindemann <oliver@expyriment.org>'
__version__ = ''
__revision__ = ''
__date__ = ''


from . import defaults
from .._internals import Expyriment_object


class Input(Expyriment_object):
    """A class implementing a general input."""

    def __init__(self):
        """Create an input."""
        Expyriment_object.__init__(self)


class Output(Expyriment_object):
    """A class implementing a general output."""

    def __init__(self):
        """Create an output."""
        Expyriment_object.__init__(self)



""" The constants module.

This module contains expyriment constants.

"""

__author__ = 'Florian Krause <florian@expyriment.org>, \
Oliver Lindemann <oliver@expyriment.org>'
__version__ = ''
__revision__ = ''
__date__ = ''


import os as _os

import pygame as _pygame

from ..misc import byte2unicode as _str2unicode
from ..misc._colour import Colour as _Colour

# Keys
K_BACKSPACE = _pygame.K_BACKSPACE
K_TAB = _pygame.K_TAB
K_CLEAR = _pygame.K_CLEAR
K_RETURN = _pygame.K_RETURN
K_PAUSE = _pygame.K_PAUSE
K_ESCAPE = _pygame.K_ESCAPE
K_SPACE = _pygame.K_SPACE
K_EXCLAIM = _pygame.K_EXCLAIM
K_QUOTEDBL = _pygame.K_QUOTEDBL
K_HASH = _pygame.K_HASH
K_DOLLAR = _pygame.K_DOLLAR
K_AMPERSAND = _pygame.K_AMPERSAND
K_QUOTE = _pygame.K_QUOTE
K_LEFTPAREN = _pygame.K_LEFTPAREN
K_RIGHTPAREN = _pygame.K_RIGHTPAREN
K_ASTERISK = _pygame.K_ASTERISK
K_PLUS = _pygame.K_PLUS
K_COMMA = _pygame.K_COMMA
K_MINUS = _pygame.K_MINUS
K_PERIOD = _pygame.K_PERIOD
K_SLASH = _pygame.K_SLASH
K_0 = _pygame.K_0
K_1 = _pygame.K_1
K_2 = _pygame.K_2
K_3 = _pygame.K_3
K_4 = _pygame.K_4
K_5 = _pygame.K_5
K_6 = _pygame.K_6
K_7 = _pygame.K_7
K_8 = _pygame.K_8
K_9 = _pygame.K_9
K_COLON = _pygame.K_COLON
K_SEMICOLON = _pygame.K_SEMICOLON
K_LESS = _pygame.K_LESS
K_EQUALS = _pygame.K_EQUALS
K_GREATER = _pygame.K_GREATER
K_QUESTION = _pygame.K_QUESTION
K_AT = _pygame.K_AT
K_LEFTBRACKET = _pygame.K_LEFTBRACKET
K_BACKSLASH = _pygame.K_BACKSLASH
K_RIGHTBRACKET = _pygame.K_RIGHTBRACKET
K_CARET = _pygame.K_CARET
K_UNDERSCORE = _pygame.K_UNDERSCORE
K_BACKQUOTE = _pygame.K_BACKQUOTE
K_a = _pygame.K_a
K_b = _pygame.K_b
K_c = _pygame.K_c
K_d = _pygame.K_d
K_e = _pygame.K_e
K_f = _pygame.K_f
K_g = _pygame.K_g
K_h = _pygame.K_h
K_i = _pygame.K_i
K_j = _pygame.K_j
K_k = _pygame.K_k
K_l = _pygame.K_l
K_m = _pygame.K_m
K_n = _pygame.K_n
K_o = _pygame.K_o
K_p = _pygame.K_p
K_q = _pygame.K_q
K_r = _pygame.K_r
K_s = _pygame.K_s
K_t = _pygame.K_t
K_u = _pygame.K_u
K_v = _pygame.K_v
K_w = _pygame.K_w
K_x = _pygame.K_x
K_y = _pygame.K_y
K_z = _pygame.K_z
K_DELETE = _pygame.K_DELETE
K_KP0 = _pygame.K_KP0
K_KP1 = _pygame.K_KP1
K_KP2 = _pygame.K_KP2
K_KP3 = _pygame.K_KP3
K_KP4 = _pygame.K_KP4
K_KP5 = _pygame.K_KP5
K_KP6 = _pygame.K_KP6
K_KP7 = _pygame.K_KP7
K_KP8 = _pygame.K_KP8
K_KP9 = _pygame.K_KP9
K_KP_PERIOD = _pygame.K_KP_PERIOD
K_KP_DIVIDE = _pygame.K_KP_DIVIDE
K_KP_MULTIPLY = _pygame.K_KP_MULTIPLY
K_KP_MINUS = _pygame.K_KP_MINUS
K_KP_PLUS = _pygame.K_KP_PLUS
K_KP_ENTER = _pygame.K_KP_ENTER
K_KP_EQUALS = _pygame.K_KP_EQUALS
K_UP = _pygame.K_UP
K_DOWN = _pygame.K_DOWN
K_RIGHT = _pygame.K_RIGHT
K_LEFT = _pygame.K_LEFT
K_INSERT = _pygame.K_INSERT
K_HOME = _pygame.K_HOME
K_END = _pygame.K_END
K_PAGEUP = _pygame.K_PAGEUP
K_PAGEDOWN = _pygame.K_PAGEDOWN
K_F1 = _pygame.K_F1
K_F2 = _pygame.K_F2
K_F3 = _pygame.K_F3
K_F4 = _pygame.K_F4
K_F5 = _pygame.K_F5
K_F6 = _pygame.K_F6
K_F7 = _pygame.K_F7
K_F8 = _pygame.K_F8
K_F9 = _pygame.K_F9
K_F10 = _pygame.K_F10
K_F11 = _pygame.K_F11
K_F12 = _pygame.K_F12
K_F13 = _pygame.K_F13
K_F14 = _pygame.K_F14
K_F15 = _pygame.K_F15
K_NUMLOCK = _pygame.K_NUMLOCK
K_CAPSLOCK = _pygame.K_CAPSLOCK
K_SCROLLOCK = _pygame.K_SCROLLOCK
K_RSHIFT = _pygame.K_RSHIFT
K_LSHIFT = _pygame.K_LSHIFT
K_RCTRL = _pygame.K_RCTRL
K_LCTRL = _pygame.K_LCTRL
K_RALT = _pygame.K_RALT
K_LALT = _pygame.K_LALT
K_RMETA = _pygame.K_RMETA
K_LMETA = _pygame.K_LMETA
K_LSUPER = _pygame.K_LSUPER
K_RSUPER = _pygame.K_RSUPER
K_MODE = _pygame.K_MODE
K_HELP = _pygame.K_HELP
K_PRINT = _pygame.K_PRINT
K_SYSREQ = _pygame.K_SYSREQ
K_BREAK = _pygame.K_BREAK
K_MENU = _pygame.K_MENU
K_POWER = _pygame.K_POWER
K_EURO = _pygame.K_EURO
K_ALL_LETTERS = list(range(K_a, K_z + 1))
K_ALL_DIGITS = list(range(K_0, K_9 + 1))
K_ALL_KEYPAD_DIGITS = list(range(K_KP0, K_KP9 + 1))

# Colours
C_BLACK = _Colour((0, 0, 0))
C_WHITE = _Colour((255, 255, 255))
C_RED = _Colour((255, 0, 0))
C_GREEN = _Colour((0, 255, 0))
C_BLUE = _Colour((0, 0, 255))
C_YELLOW = _Colour((255, 255, 0))
C_GREY = _Colour((200, 200, 200))
C_DARKGREY = _Colour((150, 150, 150))
C_EXPYRIMENT_ORANGE = _Colour((255, 150, 50))
C_EXPYRIMENT_PURPLE = _Colour((160, 70, 250))

# Permutation types
P_BALANCED_LATIN_SQUARE = 'balanced-latin-square'
P_CYCLED_LATIN_SQUARE = 'cycled-latin-square'
P_RANDOM = 'random'

# Misc
_tmp = _os.path.abspath(
    _os.path.join(_os.path.dirname(__file__),
                  "..", "expyriment_logo.png"))
EXPYRIMENT_LOGO_FILE = _str2unicode(_tmp)



"""
Audio playback.

This module contains a class implementing audio playback.

"""
from __future__ import absolute_import, print_function, division
from builtins import *

__author__ = 'Florian Krause <florian@expyriment.org>, \
Oliver Lindemann <oliver@expyriment.org>'
__version__ = ''
__revision__ = ''
__date__ = ''


import os

try:
    import android.mixer as mixer
except:
    import pygame.mixer as mixer
from .. import _internals
from ..misc import unicode2byte
from ._stimulus import Stimulus


class Audio(Stimulus):
    """A class implementing a general auditory stimulus.

    See Also
    --------
    expyriment.control.start_audiosystem
    expyriment.control.stop_audiosystem
    expyriment.control.get_audiosystem_is_playing
    expyriment.control.wait_end_audiosystem

    """

    def __init__(self, filename):
        """Create an audio stimulus.

        Parameters
        ----------
        filename : str
            the filename. Must be an .ogg or uncompressed .wav file.

        """
        if os.path.splitext(filename)[1] not in ('.wav', '.ogg'):
            raise ValueError("The audio file must be an .ogg or uncompressed .wav file")

        Stimulus.__init__(self, filename)
        self._filename = filename
        self._file = None
        self._is_preloaded = False
        if not(os.path.isfile(self._filename)):
            raise IOError(u"The audio file {0} does not exists".format(
                self._filename))

    _getter_exception_message = "Cannot set {0} if preloaded!"

    @property
    def is_preloaded(self):
        """Getter for is_preloaded."""

        return self._is_preloaded

    @property
    def filename(self):
        """Getter for filename."""

        return self._filename

    @filename.setter
    def filename(self, value):
        """Setter for filename."""

        if self._is_preloaded:
            raise AttributeError(Audio._getter_exception_message.format(
                "filename"))
        else:
            self._filename = value

    def copy(self):
        """Copy the stimulus.

        Returns
        -------
        copy: expyriment.stimuli.Audio
            Returned copy will NOT be is_preloaded!

        """

        was_loaded = self._is_preloaded
        self.unload()
        rtn = Stimulus.copy(self)
        if was_loaded:
            self.preload()
        return rtn

    def preload(self):
        """Preload stimulus to memory."""

        if not self._is_preloaded:
        # Due to a bug in handling file names in PyGame 1.9.2, we pass a file
        # handle to PyGame. See also:
        # https://github.com/expyriment/expyriment/issues/81
            with open(self._filename, 'rb') as f:
                self._file = mixer.Sound(f)
            self._is_preloaded = True

    def unload(self, **kwargs):
        """Unload stimulus from memory.

        This removes the reference to the object in memory.
        It is up to the garbage collector to actually remove it from memory.

        """

        if self._is_preloaded:
            self._file = None
            self._is_preloaded = False

    def play(self, loops=0, maxtime=0, fade_ms=0, log_event_tag=None):
        """Play the audio stimulus.

        The function returns immediately after the sound started to play.
        A pygame.mixer.Channel object is returned.

        Parameters
        ----------
        loops : int, optional
            how often to repeat (-1 = forever) (default = 0)
        maxtime : int
            stop after given amount of milliseconds (default = 0)
        fade_ms : int, optional
            fade in time in milliseconds (default = 0)
        log_event_tag : numeral or string, optional
            if log_event_tag is defined and if logging is switched on for this
            stimulus (default), a summary of the inter-event-intervalls are
            appended at the end of the event file

        """

        if not self._is_preloaded:
            self.preload()
        rtn = self._file.play(loops, maxtime, fade_ms)
        if self._logging:
            if isinstance(self._filename, str):
                import sys
                filename = self._filename.encode(sys.getfilesystemencoding())
            else:
                filename = self._filename

            _internals.active_exp._event_file_log(
                "Stimulus,played,{0}".format(filename), 1,
                                 log_event_tag=log_event_tag)
        return rtn

    def stop(self):
        """Stop the audio stimulus"""

        if self._is_preloaded:
            self._file.stop()

    def present(self, log_event_tag=None):
        """Presents the sound.

        The function is identical to Audio.play(loops=0, maxtime=0, fade_ms=0)
        and returns also immediately after the sound started to play.

        Notes
        -----
        See Audio.play for more information.

        """

        self.play(log_event_tag=log_event_tag)



"""
The statistics module.

This module contains miscellaneous stastistical functions for expyriment.

"""
from __future__ import absolute_import, print_function, division
from builtins import *

__author__ = 'Florian Krause <florian@expyriment.org>, \
Oliver Lindemann <oliver@expyriment.org>'
__version__ = ''
__revision__ = ''
__date__ = ''


import math as _math


def sum(data):
    """Returns the sum of data.

    The function ignores all non-numerical elements in the data and returns
    None if no numerical element has been found. In contrast to standard math
    and numpy functions, this function is robust against type violations.

    Parameters
    ----------
    data : list
        list of numerical data

    Returns
    -------
    out : float or None

    """

    s = 0
    elem_found = False
    for v in data:
        try:
            s += v
            elem_found = True
        except:
            pass
    if elem_found:
        return s
    else:
        return None


def mode(data):
    """Returns the mode, that is, the most frequent value in data.

    Parameters
    ----------
    data : list
        list of numerical data

    Returns
    -------
    out : float or None

    """

    freq = frequency_table(data)
    Fmax = max(freq.values())
    for x, f in list(freq.items()):
        if f == Fmax:
            break
    return x

def mean(data):
    """Returns the mean of data.

    Notes
    -----
    The function ignores all non-numerical elements in the data and returns
    None if no numerical element has been found. In contrast to standard math
    and numpy functions, this function is robust against type violations.

    Parameters
    ----------
    data : list
        list of numerical data

    Returns
    -------
    out : float or None

    """

    s = 0
    cnt = 0
    for v in data:
        try:
            s += v
            cnt += 1
        except:
            pass
    if cnt == 0:
        return None
    else:
        return float(s) / float(cnt)

def median(data):
    """Returns the median of data.

    Notes
    -----
    The function ignores all non-numerical elements in the data and returns
    None if no numerical element has been found. In contrast to standard math
    and numpy functions, this function is robust against type violations.

    Parameters
    ----------
    data : list
        list of numerical data

    Returns
    -------
    out : float or None

    """

    tmp = []
    for elem in data: # remove non numerics
        if isinstance(elem, (int, int, float)):
            tmp.append(elem)
    data = sorted(tmp)
    if len(data) % 2 == 1:
        return data[(len(data) - 1) // 2 ]
    else:
        lower = data[len(data) // 2 - 1]
        upper = data[len(data) // 2]
        return (float(lower + upper)) / 2.0

def frequence_table(data):
    """OBSOLETE FUNCTION! Please use `frequency_table`!"""
    
    raise DeprecationWarning("frequence_table is an obsolete function. Please use frequency_table.")
    
    frequency_table(data)

def frequency_table(data):
    """Returns the frequency table of the data as dictionary.

    Parameters
    ----------
    data : list
        list of numerical data

    Returns
    -------
    out : dict
        `dict.keys` : values, `dict.values` : frequencies


    """

    freq = {}
    for x in data:
        freq[x] = freq.get(x, 0) + 1
    return freq

def variance(data):
    """Returns the variance of data.

    Notes
    -----
    The function ignores all non-numerical elements in the data and returns
    None if no numerical element has been found. In contrast to standard math
    and numpy functions, this function is robust against type violations.

    Parameters
    ----------
    data : list
        list of numerical data

    Returns
    -------
    out : float or None

    """

    data_squared = map(lambda x:x**2, data)
    return mean(data_squared) - mean(data)**2

def std(data):
    """Returns the standard deviation of data.

    Notes
    -----
    The function ignores all non-numerical elements in the data and returns
    None if no numerical element has been found. In contrast to standard math
    and numpy functions, this function is robust against type violations.

    Parameters
    ----------
    data : list
        list of numerical data

    Returns
    -------
    out : float or None

    """

    return _math.sqrt(variance(data))



"""
The io extra package.

Notes
-----
    To us the extras module you have to import it manually by calling:
    `import expyriment.io.extras`

"""

__author__ = 'Florian Krause <florian@expyriment.org> \
Oliver Lindemann <oliver@expyriment.org>'
__version__ = ''
__revision__ = ''
__date__ = ''

from ... import _internals

print("IO plugins:")
for name, code in _internals.import_plugins_code("io").items():
    print(" " + name)
    try:
        exec(code)
    except Exception as err:
        print("Warning: Could not import {0}".format(name))
        print(" {0}".format(err))

try:
    del (name, code)
except:
    pass